# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains wrapper classes for convolution layers of tf.keras and Switchable Atrous Convolution.

Switchable Atrous Convolution (SAC) is convolution with a switchable atrous
rate. It also has optional pre- and post-global context layers.
[1] Siyuan Qiao, Liang-Chieh Chen, Alan Yuille. DetectoRS: Detecting Objects
    with Recursive Feature Pyramid and Switchable Atrous Convolution.
    arXiv:2006.02334
"""
from typing import Optional

import tensorflow as tf
from absl import logging
from deeplab2.model import utils
from deeplab2.model.layers import activations


def _compute_padding_size(kernel_size, atrous_rate):
    kernel_size_effective = kernel_size + (kernel_size - 1) * (atrous_rate - 1)
    pad_total = kernel_size_effective - 1
    pad_begin = pad_total // 2
    pad_end = pad_total - pad_begin
    if pad_begin != pad_end:
        logging.warn('Convolution requires one more padding to the '
                     'bottom-right pixel. This may cause misalignment.')
    return (pad_begin, pad_end)


class GlobalContext(tf.keras.layers.Layer):
    """Class for the global context modules in Switchable Atrous Convolution."""

    def build(self, input_shape):
        super().build(input_shape)
        input_shape = tf.TensorShape(input_shape)
        input_channel = self._get_input_channel(input_shape)
        self.global_average_pooling = tf.keras.layers.GlobalAveragePooling2D()
        self.convolution = tf.keras.layers.Conv2D(
            input_channel, 1, strides=1, padding='same', name=self.name + '_conv',
            kernel_initializer='zeros', bias_initializer='zeros')

    def call(self, inputs, *args, **kwargs):
        outputs = self.global_average_pooling(inputs)
        outputs = tf.expand_dims(outputs, axis=1)
        outputs = tf.expand_dims(outputs, axis=1)
        outputs = self.convolution(outputs)
        return inputs + outputs

    def _get_input_channel(self, input_shape):
        # Reference: tf.keras.layers.convolutional.Conv.
        if input_shape.dims[-1].value is None:
            raise ValueError('The channel dimension of the inputs '
                             'should be defined. Found `None`.')
        return int(input_shape[-1])


class SwitchableAtrousConvolution(tf.keras.layers.Conv2D):
    """Class for the Switchable Atrous Convolution."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._average_pool = tf.keras.layers.AveragePooling2D(
            pool_size=(5, 5), strides=1, padding='same')
        self._switch = tf.keras.layers.Conv2D(
            1,
            kernel_size=1,
            strides=self.strides,
            padding='same',
            dilation_rate=1,
            name='switch',
            kernel_initializer='zeros',
            bias_initializer='zeros')

    def large_convolution_op(self, inputs, kernel):
        if self.padding == 'causal':
            tf_padding = 'VALID'
        elif isinstance(self.padding, str):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding
        large_dilation_rate = list(self.dilation_rate)
        large_dilation_rate = [r * 3 for r in large_dilation_rate]
        return tf.nn.convolution(
            inputs, kernel,
            strides=list(self.strides),
            padding=tf_padding,
            dilations=large_dilation_rate,
            data_format=self._tf_data_format,
            name=self.__class__.__name__ + '_large'
        )

    def call(self, inputs):
        # Reference: tf.keras.layers.convolutional.Conv.
        input_shape = inputs.shape
        switches = self._switch(self._average_pool(inputs))

        if self._is_causal:  # Apply causal padding to inputs for Conv1D.
            inputs = tf.compat.v1.pad(inputs, self._compute_causal_padding(inputs))

        outputs = self.convolution_op(inputs, self.kernel)
        outputs_large = self.large_convolution_op(inputs, self.kernel)

        outputs = switches * outputs_large + (1 - switches) * outputs

        if self.use_bias:
            outputs = tf.nn.bias_add(
                outputs, self.bias, data_format=self._tf_data_format)

        if not tf.executing_eagerly():
            # Infer the static output shape:
            out_shape = self.compute_output_shape(input_shape)
            outputs.set_shape(out_shape)

        if self.activation is not None:
            return self.activation(outputs)
        return outputs

    def squeeze_batch_dims(self, inp, op, inner_rank):
        # Reference: tf.keras.utils.conv_utils.squeeze_batch_dims.
        with tf.name_scope('squeeze_batch_dims'):
            shape = inp.shape

            inner_shape = shape[-inner_rank:]
            if not inner_shape.is_fully_defined():
                inner_shape = tf.compat.v1.shape(inp)[-inner_rank:]

            batch_shape = shape[:-inner_rank]
            if not batch_shape.is_fully_defined():
                batch_shape = tf.compat.v1.shape(inp)[:-inner_rank]

            if isinstance(inner_shape, tf.TensorShape):
                inp_reshaped = tf.reshape(inp, [-1] + inner_shape.as_list())
            else:
                inp_reshaped = tf.reshape(
                    inp, tf.concat(([-1], inner_shape), axis=-1))

            out_reshaped = op(inp_reshaped)

            out_inner_shape = out_reshaped.shape[-inner_rank:]
            if not out_inner_shape.is_fully_defined():
                out_inner_shape = tf.compat.v1.shape(out_reshaped)[-inner_rank:]

            out = tf.reshape(
                out_reshaped, tf.concat((batch_shape, out_inner_shape), axis=-1))

            out.set_shape(inp.shape[:-inner_rank] + out.shape[-inner_rank:])
            return out


class Conv2DSame(tf.keras.layers.Layer):
    """A wrapper class for a 2D convolution with 'same' padding.

    In contrast to tf.keras.layers.Conv2D, this layer aligns the kernel with the
    top-left corner rather than the bottom-right corner. Optionally, a batch
    normalization and an activation can be added on top.
    """

    def __init__(
            self,
            output_channels: int,
            kernel_size: int,
            name: str,
            strides: int = 1,
            atrous_rate: int = 1,
            use_bias: bool = True,
            use_bn: bool = False,
            bn_layer: tf.keras.layers.Layer = tf.keras.layers.BatchNormalization,
            bn_gamma_initializer: str = 'ones',
            activation: Optional[str] = None,
            use_switchable_atrous_conv: bool = False,
            use_global_context_in_sac: bool = False,
            conv_kernel_weight_decay: float = 0.0):  # pytype: disable=annotation-type-mismatch  # typed-keras
        """Initializes convolution with zero padding aligned to the top-left corner.

        DeepLab aligns zero padding differently to tf.keras 'same' padding.
        Considering a convolution with a 7x7 kernel, a stride of 2 and an even input
        size, tf.keras 'same' padding will add 2 zero padding to the top-left and 3
        zero padding to the bottom-right. However, for consistent feature alignment,
        DeepLab requires an equal padding of 3 in all directions. This behavior is
        consistent with e.g. the ResNet 'stem' block.

        Args:
          output_channels: An integer specifying the number of filters of the
            convolution.
          kernel_size: An integer specifying the size of the convolution kernel.
          name: A string specifying the name of this layer.
          strides: An optional integer or tuple of integers specifying the size of
            the strides (default: 1).
          atrous_rate: An optional integer or tuple of integers specifying the
            atrous rate of the convolution (default: 1).
          use_bias: An optional flag specifying whether bias should be added for the
            convolution.
          use_bn: An optional flag specifying whether batch normalization should be
            added after the convolution (default: False).
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          bn_gamma_initializer: An initializer for the batch norm gamma weight.
          activation: An optional flag specifying an activation function to be added
            after the convolution.
          use_switchable_atrous_conv: Boolean, whether the layer uses switchable
            atrous convolution.
          use_global_context_in_sac: Boolean, whether the switchable atrous
            convolution (SAC) uses pre- and post-global context.
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.

        Raises:
          ValueError: If use_bias and use_bn in the convolution.
        """
        super(Conv2DSame, self).__init__(name=name)

        if use_bn and use_bias:
            raise ValueError('Conv2DSame is using convolution bias with batch_norm.')

        if use_global_context_in_sac:
            self._pre_global_context = GlobalContext(name='pre_global_context')

        convolution_op = tf.keras.layers.Conv2D
        convolution_padding = 'same'
        if strides == 1 or strides == (1, 1):
            if use_switchable_atrous_conv:
                convolution_op = SwitchableAtrousConvolution
        else:
            padding = _compute_padding_size(kernel_size, atrous_rate)
            self._zeropad = tf.keras.layers.ZeroPadding2D(
                padding=(padding, padding), name='zeropad')
            convolution_padding = 'valid'
        self._conv = convolution_op(
            output_channels,
            kernel_size,
            strides=strides,
            padding=convolution_padding,
            use_bias=use_bias,
            dilation_rate=atrous_rate,
            name='conv',
            kernel_initializer='he_normal',
            kernel_regularizer=tf.keras.regularizers.l2(
                conv_kernel_weight_decay))

        if use_global_context_in_sac:
            self._post_global_context = GlobalContext(name='post_global_context')

        if use_bn:
            self._batch_norm = bn_layer(axis=3, name='batch_norm',
                                        gamma_initializer=bn_gamma_initializer)

        self._activation_fn = None
        if activation is not None:
            self._activation_fn = activations.get_activation(activation)

        self._use_global_context_in_sac = use_global_context_in_sac
        self._strides = strides
        self._use_bn = use_bn

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        x = input_tensor
        if self._use_global_context_in_sac:
            x = self._pre_global_context(x)

        if not (self._strides == 1 or self._strides == (1, 1)):
            x = self._zeropad(x)
        x = self._conv(x)

        if self._use_global_context_in_sac:
            x = self._post_global_context(x)

        if self._use_bn:
            x = self._batch_norm(x, training=training)

        if self._activation_fn is not None:
            x = self._activation_fn(x)
        return x


class DepthwiseConv2DSame(tf.keras.layers.Layer):
    """A wrapper class for a 2D depthwise convolution.

    In contrast to convolutions in tf.keras.layers.DepthwiseConv2D, this layers
    aligns the kernel with the top-left corner rather than the bottom-right
    corner. Optionally, a batch normalization and an activation can be added.
    """

    def __init__(self,
                 kernel_size: int,
                 name: str,
                 strides: int = 1,
                 atrous_rate: int = 1,
                 use_bias: bool = True,
                 use_bn: bool = False,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 activation: Optional[str] = None):
        """Initializes a 2D depthwise convolution.

        Args:
          kernel_size: An integer specifying the size of the convolution kernel.
          name: A string specifying the name of this layer.
          strides: An optional integer or tuple of integers specifying the size of
            the strides (default: 1).
          atrous_rate: An optional integer or tuple of integers specifying the
            atrous rate of the convolution (default: 1).
          use_bias: An optional flag specifying whether bias should be added for the
            convolution.
          use_bn: An optional flag specifying whether batch normalization should be
            added after the convolution (default: False).
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: An optional flag specifying an activation function to be added
            after the convolution.

        Raises:
          ValueError: If use_bias and use_bn in the convolution.
        """
        super(DepthwiseConv2DSame, self).__init__(name=name)

        if use_bn and use_bias:
            raise ValueError(
                'DepthwiseConv2DSame is using convlution bias with batch_norm.')

        if strides == 1 or strides == (1, 1):
            convolution_padding = 'same'
        else:
            padding = _compute_padding_size(kernel_size, atrous_rate)
            self._zeropad = tf.keras.layers.ZeroPadding2D(
                padding=(padding, padding), name='zeropad')
            convolution_padding = 'valid'
        self._depthwise_conv = tf.keras.layers.DepthwiseConv2D(
            kernel_size=kernel_size,
            strides=strides,
            padding=convolution_padding,
            use_bias=use_bias,
            dilation_rate=atrous_rate,
            name='depthwise_conv')
        if use_bn:
            self._batch_norm = bn_layer(axis=3, name='batch_norm')

        self._activation_fn = None
        if activation is not None:
            self._activation_fn = activations.get_activation(activation)

        self._strides = strides
        self._use_bn = use_bn

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        x = input_tensor
        if not (self._strides == 1 or self._strides == (1, 1)):
            x = self._zeropad(x)
        x = self._depthwise_conv(x)
        if self._use_bn:
            x = self._batch_norm(x, training=training)
        if self._activation_fn is not None:
            x = self._activation_fn(x)
        return x


class SeparableConv2DSame(tf.keras.layers.Layer):
    """A wrapper class for a 2D separable convolution.

    In contrast to convolutions in tf.keras.layers.SeparableConv2D, this layers
    aligns the kernel with the top-left corner rather than the bottom-right
    corner. Optionally, a batch normalization and an activation can be added.
    """

    def __init__(
            self,
            output_channels: int,
            kernel_size: int,
            name: str,
            strides: int = 1,
            atrous_rate: int = 1,
            use_bias: bool = True,
            use_bn: bool = False,
            bn_layer: tf.keras.layers.Layer = tf.keras.layers.BatchNormalization,
            activation: Optional[str] = None):  # pytype: disable=annotation-type-mismatch  # typed-keras
        """Initializes a 2D separable convolution.

        Args:
          output_channels: An integer specifying the number of filters of the
            convolution output.
          kernel_size: An integer specifying the size of the convolution kernel.
          name: A string specifying the name of this layer.
          strides: An optional integer or tuple of integers specifying the size of
            the strides (default: 1).
          atrous_rate: An optional integer or tuple of integers specifying the
            atrous rate of the convolution (default: 1).
          use_bias: An optional flag specifying whether bias should be added for the
            convolution.
          use_bn: An optional flag specifying whether batch normalization should be
            added after the convolution (default: False).
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: An optional flag specifying an activation function to be added
            after the convolution.

        Raises:
          ValueError: If use_bias and use_bn in the convolution.
        """
        super(SeparableConv2DSame, self).__init__(name=name)
        if use_bn and use_bias:
            raise ValueError(
                'SeparableConv2DSame is using convolution bias with batch_norm.')

        self._depthwise = DepthwiseConv2DSame(
            kernel_size=kernel_size,
            name='depthwise',
            strides=strides,
            atrous_rate=atrous_rate,
            use_bias=use_bias,
            use_bn=use_bn,
            bn_layer=bn_layer,
            activation=activation)
        self._pointwise = Conv2DSame(
            output_channels=output_channels,
            kernel_size=1,
            name='pointwise',
            strides=1,
            atrous_rate=1,
            use_bias=use_bias,
            use_bn=use_bn,
            bn_layer=bn_layer,
            activation=activation)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        x = self._depthwise(input_tensor, training=training)
        return self._pointwise(x, training=training)


class StackedConv2DSame(tf.keras.layers.Layer):
    """Stacked Conv2DSame or SeparableConv2DSame.

    This class sequentially stacks a given number of Conv2DSame layers or
    SeparableConv2DSame layers.
    """

    def __init__(
            self,
            num_layers: int,
            conv_type: str,
            output_channels: int,
            kernel_size: int,
            name: str,
            strides: int = 1,
            atrous_rate: int = 1,
            use_bias: bool = True,
            use_bn: bool = False,
            bn_layer: tf.keras.layers.Layer = tf.keras.layers.BatchNormalization,
            activation: Optional[str] = None):  # pytype: disable=annotation-type-mismatch  # typed-keras
        """Initializes a stack of convolutions.

        Args:
          num_layers: The number of convolutions to create.
          conv_type: A string specifying the convolution type used in each block.
            Must be one of 'standard_conv' or 'depthwise_separable_conv'.
          output_channels: An integer specifying the number of filters of the
            convolution output.
          kernel_size: An integer specifying the size of the convolution kernel.
          name: A string specifying the name of this layer.
          strides: An optional integer or tuple of integers specifying the size of
            the strides (default: 1).
          atrous_rate: An optional integer or tuple of integers specifying the
            atrous rate of the convolution (default: 1).
          use_bias: An optional flag specifying whether bias should be added for the
            convolution.
          use_bn: An optional flag specifying whether batch normalization should be
            added after the convolution (default: False).
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          activation: An optional flag specifying an activation function to be added
            after the convolution.

        Raises:
          ValueError: An error occurs when conv_type is neither 'standard_conv'
            nor 'depthwise_separable_conv'.
        """
        super(StackedConv2DSame, self).__init__(name=name)
        if conv_type == 'standard_conv':
            convolution_op = Conv2DSame
        elif conv_type == 'depthwise_separable_conv':
            convolution_op = SeparableConv2DSame
        else:
            raise ValueError('Convolution %s not supported.' % conv_type)

        for index in range(num_layers):
            current_name = utils.get_conv_bn_act_current_name(index, use_bn,
                                                              activation)
            utils.safe_setattr(self, current_name, convolution_op(
                output_channels=output_channels,
                kernel_size=kernel_size,
                name=utils.get_layer_name(current_name),
                strides=strides,
                atrous_rate=atrous_rate,
                use_bias=use_bias,
                use_bn=use_bn,
                bn_layer=bn_layer,
                activation=activation))
        self._num_layers = num_layers
        self._use_bn = use_bn
        self._activation = activation

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        x = input_tensor
        for index in range(self._num_layers):
            current_name = utils.get_conv_bn_act_current_name(index, self._use_bn,
                                                              self._activation)
            x = getattr(self, current_name)(x, training=training)
        return x


class Conv1D(tf.keras.layers.Layer):
    """A wrapper class for a 1D convolution with batch norm and activation.

    Conv1D creates a convolution kernel that is convolved with the layer input
    over a single spatial (or temporal) dimension to produce a tensor of outputs.
    The input should always be 3D with shape [batch, length, channel], so
    accordingly, the optional batch norm is done on axis=2.

    In DeepLab, we use Conv1D only with kernel_size = 1 for dual path transformer
    layers in MaX-DeepLab [1] architectures.

    Reference:
    [1] MaX-DeepLab: End-to-End Panoptic Segmentation with Mask Transformers,
        CVPR 2021.
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    def __init__(
            self,
            output_channels: int,
            name: str,
            use_bias: bool = True,
            use_bn: bool = False,
            bn_layer: tf.keras.layers.Layer = tf.keras.layers.BatchNormalization,
            bn_gamma_initializer: str = 'ones',
            activation: Optional[str] = None,
            conv_kernel_weight_decay: float = 0.0,
            kernel_initializer='he_normal',
            kernel_size: int = 1,
            padding: str = 'valid'):  # pytype: disable=annotation-type-mismatch  # typed-keras
        """Initializes a Conv1D.

        Args:
          output_channels: An integer specifying the number of filters of the
            convolution.
          name: A string specifying the name of this layer.
          use_bias: An optional flag specifying whether bias should be added for the
            convolution.
          use_bn: An optional flag specifying whether batch normalization should be
            added after the convolution (default: False).
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          bn_gamma_initializer: An initializer for the batch norm gamma weight.
          activation: An optional flag specifying an activation function to be added
            after the convolution.
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.
          kernel_initializer: An initializer for the convolution kernel.
          kernel_size: An integer specifying the size of the convolution kernel.
          padding: An optional string specifying the padding to use. Must be either
            'same' or 'valid' (default: 'valid').

        Raises:
          ValueError: If use_bias and use_bn in the convolution.
        """
        super(Conv1D, self).__init__(name=name)

        if use_bn and use_bias:
            raise ValueError('Conv1D is using convlution bias with batch_norm.')

        self._conv = tf.keras.layers.Conv1D(
            output_channels,
            kernel_size=kernel_size,
            strides=1,
            padding=padding,
            use_bias=use_bias,
            name='conv',
            kernel_initializer=kernel_initializer,
            kernel_regularizer=tf.keras.regularizers.l2(
                conv_kernel_weight_decay))

        self._batch_norm = None
        if use_bn:
            # Batch norm uses axis=2 because the input is 3D with channel being the
            # last dimension.
            self._batch_norm = bn_layer(axis=2, name='batch_norm',
                                        gamma_initializer=bn_gamma_initializer)

        self._activation_fn = None
        if activation is not None:
            self._activation_fn = activations.get_activation(activation)

    def call(self, input_tensor, training=False):
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, length,
            channels].
          training: A boolean flag indicating whether training behavior should be
            used (default: False).

        Returns:
          The output tensor.
        """
        x = self._conv(input_tensor)
        if self._batch_norm is not None:
            x = self._batch_norm(x, training=training)
        if self._activation_fn is not None:
            x = self._activation_fn(x)
        return x



from .torchview import draw_graph
from .computation_graph import ComputationGraph
from .computation_node import Node, TensorNode, ModuleNode, FunctionNode
from .recorder_tensor import RecorderTensor

__all__ = (
    "draw_graph",
    'ComputationGraph',
    'Node',
    'FunctionNode',
    'ModuleNode',
    'TensorNode',
    'RecorderTensor',
)
__version__ = "0.2.6"



"""
This is a very task-specific file containing logic to parse fusion pipeline
metrics for BAS and SC.

Used by ./aggregate_loader.py
"""

import numpy as np
import warnings
import json
import ubelt as ub
import io
# import xdev
import pandas as pd
import re
from kwutil import util_time
from kwutil import util_pattern


# Do we need to memoize this?
def parse_json_header(fpath):
    """
    Ideally the information we need is in the first few bytes of the json file
    """
    from geowatch.utils import ijson_ext
    import zipfile
    if zipfile.is_zipfile(fpath):
        # We have a compressed json file, but we can still read the header
        # fairly quickly.
        zfile = zipfile.ZipFile(fpath)
        names = zfile.namelist()
        assert len(names) == 1
        member = names[0]
        # Stream the header directly from the zipfile.
        file = zfile.open(member, 'r')
    else:
        # Normal json file
        file = open(fpath, 'r')

    with file:
        # import ijson
        # We only expect there to be one info section
        # try:
        #     # Try our extension if the main library fails (due to NaN)
        #     info_section_iter = ijson.items(file, prefix='info')
        #     info_section = next(info_section_iter)
        # except ijson.IncompleteJSONError:
        # Try our extension if the main library fails (due to NaN)
        # file.seek(0)

        # Nans are too frequent, only use our extension
        info_section_iter = ijson_ext.items(file, prefix='info')
        info_section = next(info_section_iter)
    return info_section

parse_json_header_cached = ub.memoize(parse_json_header)


def trace_json_lineage(fpath):
    """
    We will expect a json file to contain a top-level "info" section that
    indicates how it is derived.

    fpath = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/fusion/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/pred/trk/Drop4_BAS_Retrain_V002_epoch=31-step=16384.pt.pt/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC_data.kwcoco/trk_pxl_b788335d/trk_poly_f2218f0b/tracks.kwcoco.json'
    """

    info_section = parse_json_header(fpath)

    # TODO:
    # uniqify by uuid
    for proc in list(find_info_items(info_section, {'process', 'process_context'})):
        name = proc['properties']['name']
        if name not in {'coco_align_geotiffs', 'coco_align'}:
            print(f'name={name}')
            print(proc['properties']['start_timestamp'])
            print(proc['properties']['emissions']['run_id'])
            print('proc = {}'.format(ub.urepr(proc, nl=2)))
        # print(proc['properties']['name'])


# def trace_kwcoco_lineage(fpath):
def load_iarpa_evaluation(fpath):
    """
    Args:
        fpath (PathLike | str):
            path to the IARPA summary json file

    Returns:
        Dict: containing keys:
            metrics -
                which just contains a flat Dict[str, float] metric dictionary
            iarpa_info -
                which contains ALL of the information parsed out of the summary json file.

    Ignore:
        fpath = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/fusion/Drop4-BAS/eval/trk/package_epoch0_step41.pt.pt/Drop4-BAS_KR_R001.kwcoco/trk_pxl_fd9e1a95/trk_poly_9f08fb8c/merged/summary2.json'
    """
    iarpa_json = _load_json(fpath)
    metrics = {}
    unique_regions = set()
    if 'best_bas_rows' in iarpa_json:
        best_bas_rows = pd.read_json(
            io.StringIO(json.dumps(iarpa_json['best_bas_rows'])),
            orient='table')
        bas_row = best_bas_rows.loc['__macro__'].reset_index().iloc[0]

        metrics.update({
            'bas_tp': bas_row['tp sites'],
            'bas_fp': bas_row['fp sites'],
            'bas_fn': bas_row['fn sites'],
            'bas_ntrue': bas_row['total sites'],
            'bas_npred': bas_row['proposed slices'],
            'bas_ppv': bas_row['precision'],
            'bas_tpr': bas_row['recall (PD)'],
            'bas_ffpa': bas_row['ffpa'],
            'bas_f1': bas_row['F1'],
            'rho': bas_row['rho'],
            'tau': bas_row['tau'],
            'bas_space_FAR': bas_row['spatial FAR'],
            'bas_time_FAR': bas_row['temporal FAR'],
            'bas_image_FAR': bas_row['images FAR'],
        })
        alpha = 1.0
        metrics['bas_faa_f1'] = metrics['bas_f1'] * (1 - metrics['bas_ffpa']) ** alpha

        unique_regions.update(
            best_bas_rows.index.levels[best_bas_rows.index.names.index('region_id')].difference({'__macro__', '__micro__'})
        )

    if 'sc_df' in iarpa_json:
        sc_json_data = iarpa_json['sc_df']
        sc_json_text = json.dumps(sc_json_data)
        try:
            sc_df = pd.read_json(io.StringIO(sc_json_text), orient='table')
        except pd.errors.IntCastingNaNError:
            # This seems like a pandas bug. It looks like it can save a Int64
            # with NaN exteions, but it can't load it back in.
            sc_json_data = iarpa_json['sc_df']
            walker = ub.IndexableWalker(sc_json_data)
            for path, val in walker:
                if path[-1] == 'extDtype' and val == 'Int64':
                    walker[path] = 'number'
                if path[-1] == 'type' and val == 'integer':
                    walker[path] = 'number'
            sc_json_text = json.dumps(sc_json_data)
            sc_df = pd.read_json(io.StringIO(sc_json_text), orient='table')

        site_prep_f1 = sc_df.loc['__macro__', 'Site Preparation']['F1']
        active_f1 = sc_df.loc['__macro__', 'Active Construction']['F1']

        site_prep_te = sc_df.loc['__macro__', 'Site Preparation']['TE']
        active_te = sc_df.loc['__macro__', 'Active Construction']['TE']
        post_te = sc_df.loc['__macro__', 'Post Construction']['TE']

        metrics.update({
            # 'mean_f1': sc_df.loc['F1'].mean(),
            'sc_macro_f1': (site_prep_f1 + active_f1) / 2,
            'macro_f1_siteprep': site_prep_f1,
            'macro_f1_active': active_f1,

            'sc_macro_te': sum([site_prep_te, active_te, post_te]) / 3,
            'macro_te_siteprep': site_prep_te,
            'macro_te_active': active_te,
            'macro_te_post': post_te,
        })

        if '__micro__' in sc_df.index:
            # Not sure why micro sometimes is not included.
            metrics['sc_micro_f1'] = sc_df.loc['__micro__']['F1'].mean()
        else:
            metrics['sc_micro_f1'] = np.nan

        unique_regions.update(
            sc_df.index.levels[sc_df.index.names.index('region_id')].difference({'__macro__', '__micro__'})
        )

    # quick and dirty way to get access to single-region results
    region_ids = ','.join(sorted(unique_regions))
    iarpa_json['region_ids'] = region_ids

    iarpa_result = {
        'metrics': metrics,
        'iarpa_json': iarpa_json,
    }
    return iarpa_result


def _handle_process_item(item):
    """
    Json data written by the process context has changed over time slightly.
    Consolidate different usages until a consistent API and usage patterns are
    established.

    """
    assert item['type'] in {'process', 'process_context'}
    props = item['properties']

    needs_modify = 0

    config = props.get('config', None)
    args = props.get('args', None)
    if config is None:
        # Use args if config is not available
        config = args
        needs_modify = True

    FIX_BROKEN_SCRIPTCONFIG_HANDLING = 1
    if FIX_BROKEN_SCRIPTCONFIG_HANDLING:
        if '_data' in config:
            config = config['_data']
            needs_modify = True
        if '_data' in args:
            args = args['_data']
            needs_modify = True

    assert 'pred_info' not in item, 'should be in extra instead'

    if needs_modify:
        import copy
        item = copy.deepcopy(item)
        item['properties']['config'] = config
        item['properties']['args'] = args

    return item


def load_pxl_eval(fpath, expt_dvc_dpath=None, arg_prefix='', mode=0, with_param_types=True):
    from kwcoco.coco_evaluator import CocoSingleResult
    # from geowatch.utils import result_analysis
    # from kwutil import util_time
    measure_info = _load_json(fpath)

    # meta = measure_info['meta']
    # pred_info = meta['info']
    # dvc_dpath = expt_dvc_dpath

    salient_measures = measure_info['nocls_measures']
    class_measures = measure_info['ovr_measures']

    # HACK: fixme
    coi_pattern = util_pattern.MultiPattern.coerce(
        ['Active Construction', 'Site Preparation'], hint='glob')

    class_metrics = []
    coi_metrics = []
    for catname, bin_measure in class_measures.items():
        class_row = {}
        class_row['AP'] = bin_measure['ap']
        class_row['AUC'] = bin_measure['auc']
        class_row['APUC'] = np.nanmean([bin_measure['ap'], bin_measure['auc']])
        class_row['catname'] = catname
        if coi_pattern.match(catname):
            coi_metrics.append(class_row)
        class_metrics.append(class_row)

    class_aps = [r['AP'] for r in class_metrics]
    class_aucs = [r['AUC'] for r in class_metrics]
    coi_aps = [r['AP'] for r in coi_metrics]
    coi_aucs = [r['AUC'] for r in coi_metrics]
    coi_catnames = [r['catname'] for r in coi_metrics]

    metrics = {}
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'Mean of empty slice')
        metrics['class_mAP'] = np.nanmean(class_aps) if len(class_aps) else np.nan
        metrics['class_mAUC'] = np.nanmean(class_aucs) if len(class_aucs) else np.nan
        metrics['class_mAPUC'] = np.nanmean([metrics['class_mAUC'], metrics['class_mAP']])

        metrics['coi_mAP'] = np.nanmean(coi_aps) if len(coi_aps) else np.nan
        metrics['coi_mAUC'] = np.nanmean(coi_aucs) if len(coi_aucs) else np.nan
        metrics['coi_mAPUC'] = np.nanmean([metrics['coi_mAUC'], metrics['coi_mAP']])

        metrics['salient_AP'] = salient_measures['ap']
        metrics['salient_AUC'] = salient_measures['auc']
        metrics['salient_APUC'] = np.nanmean([metrics['salient_AP'], metrics['salient_AUC']])

    for class_row in class_metrics:
        metrics[class_row['catname'] + '_AP'] = class_row['AP']
        metrics[class_row['catname'] + '_AUC'] = class_row['AUC']

    result = CocoSingleResult.from_json(measure_info)
    json_info = measure_info.copy()

    extra_attrs = {}
    param_types = None

    info = {
        'fpath': fpath,
        'metrics': metrics,
        'param_types': param_types,
        'other': {
            'result': result,
            'extra_attrs': extra_attrs,
            'coi_catnames': ','.join(sorted(coi_catnames)),
            # 'sc_cm': sc_cm,
            # 'sc_df': sc_df,
        },
        'json_info': json_info,
    }

    return info


class Found(Exception):
    pass


def resolve_cross_machine_path(path, dvc_dpath=None):
    """
    HACK

    Attempt to determine what the local path to a file/directry would be
    if it exists on this machine. This assumes the path is something
    that was checked into DVC.

    Args:
        dvc_dpath : the preferred dvc dpath to associate the file with
            in case the older one points to multiple.
    """
    # pkg_dvc = SimpleDVC.find_root(package_fpath).resolve()
    # SimpleDVC.find_root(package_fpath).resolve()
    path = ub.Path(path)
    needs_resolve = not path.exists()
    if not needs_resolve:
        if dvc_dpath is not None:
            needs_resolve = not path.is_relative_to(dvc_dpath)

    if needs_resolve:
        expected_dnames = [
            'smart_watch_dvc',
            'smart_watch_dvc-hdd',
            'smart_watch_dvc-ssd',
        ]

        found_idx = None
        for dname in expected_dnames:
            try:
                idx = path.parts.index(dname)
            except ValueError:
                pass
            else:
                found_idx = idx
                break

        if found_idx is not None:
            # import geowatch
            # dvc_dpath = geowatch.find_dvc_dpath()
            pname = ub.Path(*path.parts[idx + 1:])
            pname_dvc = pname.augment(tail='.dvc')
            cwd = ub.Path('.').absolute()
            candidates = []
            if dvc_dpath is not None:
                candidates.extend([
                    dvc_dpath / pname,
                    dvc_dpath / pname_dvc,
                ])
            candidates.extend([
                cwd / pname,
                cwd / pname_dvc,
            ])
            found = None
            try:
                for cand_path in candidates:
                    if cand_path.exists():
                        found = cand_path
                        if found.name.endswith('.dvc'):
                            found = found.augment(ext='')
                        raise Found
            except Found:
                pass
            if found:
                return found
    return path


@ub.memoize
def global_ureg():
    import pint
    ureg = pint.UnitRegistry()
    return ureg


global_ureg()


def _add_prefix(prefix, dict_):
    return {prefix + k: v for k, v in dict_.items()}


def relevant_pred_pxl_config(pred_pxl_config, dvc_dpath=None, arg_prefix=''):
    # TODO: better way of inferring what params are relevant
    # This should be metadata a scriptconfig object can hold.
    pred_config = {}
    pred_config['tta_fliprot'] = pred_pxl_config.get('tta_fliprot', 0)
    pred_config['tta_time'] = pred_pxl_config.get('tta_time', 0)
    pred_config['chip_overlap'] = pred_pxl_config['chip_overlap']
    pred_config['input_space_scale'] = pred_pxl_config.get('input_space_scale', None)
    pred_config['window_space_scale'] = pred_pxl_config.get('window_space_scale', None)
    pred_config['output_space_scale'] = pred_pxl_config.get('output_space_scale', None)
    pred_config['time_span'] = pred_pxl_config.get('time_span', None)
    pred_config['time_sampling'] = pred_pxl_config.get('time_sampling', None)
    pred_config['time_steps'] = pred_pxl_config.get('time_steps', None)
    pred_config['chip_dims'] = pred_pxl_config.get('chip_dims', None)
    pred_config['set_cover_algo'] = pred_pxl_config.get('set_cover_algo', None)
    pred_config['resample_invalid_frames'] = pred_pxl_config.get('resample_invalid_frames', None)
    pred_config['use_cloudmask'] = pred_pxl_config.get('use_cloudmask', None)
    package_fpath = pred_pxl_config['package_fpath']
    test_dataset = pred_pxl_config['test_dataset']
    if dvc_dpath is not None:
        package_fpath = resolve_cross_machine_path(package_fpath, dvc_dpath)
        test_dataset = resolve_cross_machine_path(test_dataset, dvc_dpath)

    # pred_config['model_fpath'] = package_fpath
    # pred_config['in_dataset'] = test_dataset
    pred_config['package_fpath'] = package_fpath
    pred_config['test_dataset'] = test_dataset

    # FIXME: use a common heuristic to transform a path into a model.
    test_dataset = ub.Path(test_dataset)
    pred_config['properties.model_name'] = model_name = ub.Path(package_fpath).name
    pred_config['properties.dataset_name'] = str(ub.Path(*test_dataset.parts[-2:]))

    # Hack to get the epoch/step/expt_name
    try:
        epoch = int(model_name.split('epoch=')[1].split('-')[0])
    except Exception:
        epoch = -1
    try:
        step = int(model_name.split('step=')[1].split('-')[0])
    except Exception:
        step = -1
    try:
        expt_name = model_name.split('_epoch=')[0]
    except Exception:
        expt_name = '?'
        # expt_name = predict_args[expt_name]

    pred_config['properties.step'] = step
    pred_config['properties.epoch'] = epoch
    pred_config['properties.expt_name'] = expt_name
    pred_config = _add_prefix(arg_prefix + 'pxl.', pred_config)
    return pred_config


def parse_resource_item(item, arg_prefix='', add_prefix=True):
    resources = {}

    ureg = global_ureg()
    pred_prop = item['properties']

    start_time = util_time.coerce_datetime(pred_prop.get('start_timestamp', None))
    end_time = util_time.coerce_datetime(pred_prop.get('end_timestamp', pred_prop.get('stop_timestamp', None)))
    iters_per_second = pred_prop.get('iters_per_second', None)
    if start_time is None or end_time is None:
        total_hours = None
    else:
        total_hours = (end_time - start_time).total_seconds() / (60 * 60)
    resources['total_hours'] = total_hours
    if iters_per_second is not None:
        resources['iters_per_second'] = iters_per_second

    if 'duration' in pred_prop:
        resources['duration'] = pred_prop['duration']

    try:
        vram = pred_prop['device_info']['allocated_vram']
        vram_gb = ureg.parse_expression(f'{vram} bytes').to('gigabytes').m
        resources['vram_gb'] = vram_gb
    except KeyError:
        ...

    hardware_parts = []

    if 'machine' in pred_prop:
        cpu_name = pred_prop['machine']['cpu_brand']
        cpu_name = re.sub('.*Gen Intel.R. Core.TM. ', '', cpu_name)
        resources['cpu_name'] = cpu_name
        hardware_parts.append(cpu_name)

    try:
        gpu_name = pred_prop['device_info']['device_name']
        resources['gpu_name'] = gpu_name
        hardware_parts.append(gpu_name)
    except KeyError:
        ...

    if 'emissions' in pred_prop:
        co2_kg = pred_prop['emissions']['co2_kg']
        kwh = pred_prop['emissions']['total_kWH']
        resources['co2_kg'] = co2_kg
        resources['kwh'] = kwh

    if 'disk_info' in pred_prop:
        disk_type = pred_prop['disk_info']['filesystem']
        resources['disk_type'] = disk_type

    resources['hardware'] = ' '.join(hardware_parts)
    if add_prefix:
        resources = _add_prefix(arg_prefix + 'resource.', resources)
    return resources


def find_pred_pxl_item(pred_info):
    pred_items = list(find_info_items(
        pred_info,
        {'process', 'process_context'},
        'geowatch.tasks.fusion.predict'
    ))
    assert len(pred_items) == 1
    pred_item = pred_items[0]
    return pred_item


def find_info_items(info, query_type, query_name=None):
    from kwutil import util_pattern
    if query_name is None:
        query_name = '*'
    query_name_pattern = util_pattern.MultiPattern.coerce(query_name)
    query_type_pattern = util_pattern.MultiPattern.coerce(query_type)
    for item in info:
        if query_type_pattern.match(item['type']):
            name = item['properties']['name']
            if query_name_pattern.match(name):
                yield item


# @ub.memoize
def _load_json(fpath):
    # memo hack for development
    with open(fpath, 'r') as file:
        data = json.load(file)
    return data


def find_track_item(tracker_info):
    tracker_alias = {
        'watch.cli.run_tracker',
        'geowatch.cli.run_tracker',
    }
    track_items = list(find_info_items(
        tracker_info,
        {'process', 'process_context'},
        tracker_alias
    ))
    if len(track_items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(track_items)}
            '''))
    track_item = track_items[0]
    return track_item


def find_metrics_framework_item(info):
    task_aliases = {
        'geowatch.cli.run_metrics_framework',
    }
    items = list(find_info_items(
        info,
        {'process', 'process_context'},
        task_aliases
    ))
    if len(items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(items)}
            '''))
    item = items[0]
    return item


def find_pxl_eval_item(info):
    task_aliases = {
        'geowatch.tasks.fusion.evaluate',
    }
    items = list(find_info_items(
        info,
        {'process', 'process_context'},
        task_aliases
    ))
    if len(items) != 1:
        raise AssertionError(ub.paragraph(
            f'''
            We should be able to find exactly 1 tracker process item,
            but instead we found {len(items)}
            '''))
    item = items[0]
    return item



"""
Handle the batch transfer of COLD features to time averaged data

Ignore:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=hdd)
    python ~/code/watch/geowatch/cli/queue_cli/prepare_cold_transfer.py \
        --src_kwcocos "$DVC_DATA_DPATH/Aligned-Drop7/*/*cold.kwcoco.zip" \
        --dst_kwcocos "$DVC_DATA_DPATH/Drop7-MedianNoWinter10GSD/*_I2LS.kwcoco.zip" \
        --run=1


    geowatch stats combo_imganns-KR_R001_I2LSC.kwcoco.zip

    geowatch visualize \
        combo_imganns-KR_R001_I2LSC.kwcoco.zip \
        --smart=True \
        --channels="(red|green|blue,pan,red_COLD_a1|green_COLD_a1|blue_COLD_a1,red_COLD_cv|green_COLD_cv|blue_COLD_cv,red_COLD_rmse|green_COLD_rmse|blue_COLD_rmse,sam.0:3,landcover_hidden.0:3,invariants.0:3)"

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=hdd)
    python -m geowatch.cli.queue_cli.prepare_splits \
        --base_fpath=$DVC_DATA_DPATH/Drop7-MedianNoWinter10GSD/combo_imganns*_I2LSC*.kwcoco.zip \
        --suffix=I2LSC \
        --backend=tmux --tmux_workers=6 \
        --run=1
"""
#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub
from cmd_queue.cli_boilerplate import CMDQueueConfig


class PrepareColdTransferConfig(CMDQueueConfig):
    """
    Run geowatch.tasks.cold.transfer_features on multiple regions in a cmd-queue
    """
    src_kwcocos = scfg.Value(None, help='input pattern for cold kwcoco files')
    dst_kwcocos = scfg.Value(None, help='pattern for cold files to transfer onto. Note this is *not* the output')
    new_suffix = scfg.Value('C', help='the suffix feature char code to append at the end of the coco name. Defaults to C for COLD')

    skip_existing = scfg.Value(True)
    cache = scfg.Value(True)


def main(cmdline=1, **kwargs):
    """
    Example:
        >>> # xdoctest: +SKIP
        >>> cmdline = 0
        >>> kwargs = dict()
        >>> kwargs = {
        >>>     'src_kwcocos': '/home/joncrall/remote/Ooo/data/dvc-repos/smart_data_dvc/Aligned-Drop7/*/*cold.kwcoco.zip',
        >>>     'dst_kwcocos': '/home/joncrall/remote/Ooo/data/dvc-repos/smart_data_dvc/Drop7-MedianNoWinter10GSD/*_I2L.kwcoco.zip',
        >>> }
        >>> main(cmdline=cmdline, **kwargs)
    """
    import rich
    from kwutil import util_path
    config = PrepareColdTransferConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    rich.print('config = ' + ub.urepr(config, nl=1))

    # Heuristic to extract region ids from file paths. This is not robust.
    import re
    region_id_pat = r'(?P<region_id>[A-Z]{2}_[CRT]\d{3})'
    prefix_pat = r'(?P<prefix>.*)'
    suffix_pat = r'(?P<suffix>.*)'
    region_coco_name = re.compile(prefix_pat + region_id_pat + suffix_pat + '.kwcoco.zip')

    # Find assocation between src and dst paths based on heuristic region-id
    src_fpaths = util_path.coerce_patterned_paths(config.src_kwcocos)
    dst_fpaths = util_path.coerce_patterned_paths(config.dst_kwcocos)

    region_id_to_src_paths = ub.ddict(list)
    region_id_to_dst_paths = ub.ddict(list)
    for fpath in src_fpaths:
        match = region_coco_name.match(fpath.name)
        region_id = match.groupdict()['region_id']
        region_id_to_src_paths[region_id].append(fpath)

    for fpath in dst_fpaths:
        match = region_coco_name.match(fpath.name)
        region_id = match.groupdict()['region_id']
        region_id_to_dst_paths[region_id].append(fpath)

    # Given the association, the reset of this script is robust.
    assert max(map(len, region_id_to_src_paths.values())) == 1
    assert max(map(len, region_id_to_dst_paths.values())) == 1

    common_region_ids = set(region_id_to_src_paths) & set(region_id_to_dst_paths)

    from geowatch.mlops.pipeline_nodes import ProcessNode

    def submit_job_step(node, depends=None, name=None):
        if config.skip_existing and node.outputs_exist:
            job = None
        else:
            node.cache = config.cache
            job = queue.submit(node.final_command(), depends=depends, name=name)
        return job

    queue = config.create_queue()
    for region_id in common_region_ids:
        src_fpath = region_id_to_src_paths[region_id][0]
        dst_fpath = region_id_to_dst_paths[region_id][0]
        match = region_coco_name.match(dst_fpath.name)
        prefix = match.groupdict()['prefix']
        suffix = match.groupdict()['suffix']
        if suffix == '':
            suffix = '_'
        suffix = suffix + config.new_suffix
        new_fname = prefix + region_id + suffix + '.kwcoco.zip'
        new_fpath = dst_fpath.parent / new_fname

        node = ProcessNode(
            command=ub.codeblock(
                r'''
                python -m geowatch.tasks.cold.transfer_features
                '''),
            in_paths={
                'coco_fpath': src_fpath,
                'combine_fpath': dst_fpath,
            },
            out_paths={
                'new_coco_fpath': new_fpath,
            },
            config={
                'copy_assets': True,
                'io_workers': 4,
            }
        )
        submit_job_step(node, name=f'transfer-cold-{region_id}')

    config.run_queue(queue)


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/geowatch/cli/queue_cli/prepare_cold_transfer.py
        python -m geowatch.cli.queue_cli.prepare_cold_transfer
    """
    main()



import os
import numpy as np
from skimage import io
from scipy.ndimage import zoom


def adjust_shape(Im, s):
    "Adjust shape of grayscale image Im to s."

    # crop if necesary
    Im = Im[:s[0], :s[1]]
    si = Im.shape

    # pad if necessary
    p0 = max(0, s[0] - si[0])
    p1 = max(0 , s[1] - si[1])

    return np.pad(Im, ((0, p0), (0, p1)), 'edge')


def read_sentinel_img(path, normalize=False):
    """Read cropped Sentinel-2 image: RGB bands."""
    im_name = os.listdir(path)[0][:-7]
    r = io.imread(path + im_name + "B04.tif")
    g = io.imread(path + im_name + "B03.tif")
    b = io.imread(path + im_name + "B02.tif")

    Im = np.stack((r, g, b), axis=2).astype('float32')

    return Im


def read_sentinel_img_4(path, normalize=False):
    """Read cropped Sentinel-2 image: RGB and NIR bands."""
    im_name = os.listdir(path)[0][:-7]
    r = io.imread(path + im_name + "B04.tif")
    g = io.imread(path + im_name + "B03.tif")
    b = io.imread(path + im_name + "B02.tif")
    nir = io.imread(path + im_name + "B08.tif")

    Im = np.stack((b, g, r, nir), axis=2).astype('float32')

    return Im


def read_sentinel_img_leq20(path, normalize=False):
    """Read cropped Sentinel-2 image: bands with resolution less than or equals to 20m."""
    im_name = os.listdir(path)[0][:-7]

    r = io.imread(path + im_name + "B04.tif")
    g = io.imread(path + im_name + "B03.tif")
    b = io.imread(path + im_name + "B02.tif")
    nir = io.imread(path + im_name + "B08.tif")
    s = r.shape

    ir1 = adjust_shape(zoom(io.imread(path + im_name + "B05.tif"), 2), s)
    ir2 = adjust_shape(zoom(io.imread(path + im_name + "B06.tif"), 2), s)
    ir3 = adjust_shape(zoom(io.imread(path + im_name + "B07.tif"), 2), s)
    nir2 = adjust_shape(zoom(io.imread(path + im_name + "B8A.tif"), 2), s)
    swir2 = adjust_shape(zoom(io.imread(path + im_name + "B11.tif"), 2), s)
    swir3 = adjust_shape(zoom(io.imread(path + im_name + "B12.tif"), 2), s)

    Im = np.stack((b, g, r, ir1, ir2, ir3, nir, nir2, swir2, swir3), axis=2).astype('float32')

    return Im


def read_sentinel_img_leq60(path, normalize=False):
    """Read cropped Sentinel-2 image: all bands."""
    im_name = os.listdir(path)[0][:-7]

    r = io.imread(path + im_name + "B04.tif")
    s = r.shape
    g = io.imread(path + im_name + "B03.tif")
    b = io.imread(path + im_name + "B02.tif")
    nir = io.imread(path + im_name + "B08.tif")

    ir1 = adjust_shape(zoom(io.imread(path + im_name + "B05.tif"), 2), s)
    ir2 = adjust_shape(zoom(io.imread(path + im_name + "B06.tif"), 2), s)
    ir3 = adjust_shape(zoom(io.imread(path + im_name + "B07.tif"), 2), s)
    nir2 = adjust_shape(zoom(io.imread(path + im_name + "B8A.tif"), 2), s)
    swir2 = adjust_shape(zoom(io.imread(path + im_name + "B11.tif"), 2), s)
    swir3 = adjust_shape(zoom(io.imread(path + im_name + "B12.tif"), 2), s)

    uv = adjust_shape(zoom(io.imread(path + im_name + "B01.tif"), 6), s)
    wv = adjust_shape(zoom(io.imread(path + im_name + "B09.tif"), 6), s)
    swirc = adjust_shape(zoom(io.imread(path + im_name + "B10.tif"), 6), s)

    Im = np.stack((uv, b, g, r, ir1, ir2, ir3, nir, nir2, wv, swirc, swir2, swir3), axis=2).astype('float32')

    return Im


def read_sentinel_img_trio(img_path, mask_path, num_channels=13, normalize=False):
    """Read cropped Sentinel-2 image pair and change map."""
    #read images
    if num_channels == 3:
        I1 = read_sentinel_img(img_path + '/imgs_1/', normalize)
        I2 = read_sentinel_img(img_path + '/imgs_2/', normalize)
    elif num_channels == 4:
        I1 = read_sentinel_img_4(img_path + '/imgs_1/', normalize)
        I2 = read_sentinel_img_4(img_path + '/imgs_2/', normalize)
    elif num_channels == 10:
        I1 = read_sentinel_img_leq20(img_path + '/imgs_1/', normalize)
        I2 = read_sentinel_img_leq20(img_path + '/imgs_2/', normalize)
    elif num_channels == 13:
        I1 = read_sentinel_img_leq60(img_path + '/imgs_1/', normalize)
        I2 = read_sentinel_img_leq60(img_path + '/imgs_2/', normalize)
    cm = io.imread(mask_path + '/cm/cm.png', as_gray=True) != 0

    # crop if necessary
    s1 = I1.shape
    s2 = I2.shape
    I2 = np.pad(I2, ((0, s1[0] - s2[0]), (0, s1[1] - s2[1]), (0, 0)), 'edge')

    return I1, I2, cm



class FilterEmptyPolygons(object):
    """Removes None elements of input data_list"""

    def __init__(self, key):
        self.key = key

    def _filter(self, item):
        if item[self.key] is not None:
            return len(item[self.key])
        else:
            return False

    def __call__(self, data_list):
        return [item for item in data_list if self._filter(item)]



#!/usr/bin/env python3
"""
Handles building datasets for AC/SC
"""
from geowatch.cli.smartflow_ingress import smartflow_ingress
from geowatch.cli.smartflow_egress import smartflow_egress
from geowatch.utils.util_framework import download_region
import ubelt as ub
import scriptconfig as scfg


class ACSCDatasetConfig(scfg.DataConfig):
    """
    Generate cropped KWCOCO dataset for SC
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='DEPRECATD. DO NOT USE')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
        '''
        Output as simple newline separated STAC items

        UNUSED.
        '''))
    jobs = scfg.Value(1, type=int, short_alias=['j'], help='DO NOT USE. SET WORKERS IN acsc_align_config instead')
    dont_recompute = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        Will not recompute if output_path already exists
        '''))

    acsc_cluster_config = scfg.Value(None, help=ub.paragraph(
        '''
        The configuration for the site-cluster-step passed to
        geowatch.cli.cluster_sites.
        If None, then no site clustering is done, and a kwcoco video is
        produced for each individual site.
        '''))

    acsc_align_config = scfg.Value(None, help=ub.paragraph(
        '''
        The configuration for the coco-align step
        '''))

    input_region_models_asset_name = scfg.Value('sv_out_region_models', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to use as input
        '''))

    input_site_models_asset_name = scfg.Value('sv_out_site_models', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to to use as input
        '''))


def main():
    config = ACSCDatasetConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))
    run_generate_acsc_cropped_kwcoco(config)


def run_generate_acsc_cropped_kwcoco(config):
    from kwutil.util_yaml import Yaml
    from geowatch.utils import util_framework
    from geowatch.cli import coco_align
    from geowatch.utils import util_fsspec
    from geowatch.mlops.pipeline_nodes import ProcessNode

    if config.aws_profile is not None:
        # This should be sufficient, but it is not tested.
        util_fsspec.S3Path._new_fs(profile=config.aws_profile)

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    if config.dont_recompute:
        output_path = util_fsspec.FSPath.coerce(config.output_path)
        if output_path.exists():
            # If output_path file was there, nothing to do
            return

    ingress_dir = ub.Path('/tmp/ingress')
    local_region_path = ub.Path('/tmp/region.json')
    node_state.print_current_state(ingress_dir)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")

    # TODO:
    # Compute kwcoco-for-sc here rather than in run-bas-datagen
    ingressed_assets = smartflow_ingress(
        input_path=config.input_path,
        assets=[
            'kwcoco_for_sc',
            config.input_region_models_asset_name,  # 'sv_out_region_models',
            # 'cropped_region_models_bas'
        ],
        outdir=ingress_dir,
        aws_profile=config.aws_profile,
        dryrun=config.dryrun,
        dont_error_on_missing_asset=True
    )

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = download_region(config.input_region_path,
                                        local_region_path,
                                        aws_profile=config.aws_profile,
                                        strip_nonregions=True)

    # Parse region_id from original region file
    region_id = util_framework.determine_region_id(local_region_path)
    print(f'region_id={region_id}')

    if region_id is None:
        raise RuntimeError("Couldn't parse 'region_id' from input region file")

    # Paths to inputs generated in previous pipeline steps
    if config.input_region_models_asset_name in ingressed_assets:
        input_region_path = ub.Path(ingressed_assets[config.input_region_models_asset_name]) / f'{region_id}.geojson'
    else:
        raise AssertionError("Just pass the right key to input_region_models_asset_name")
        # print("* Didn't find region output from SV; using region output "
        #       "from BAS *")
        # input_region_path = ub.Path(ingressed_assets['cropped_region_models_bas']) / f'{region_id}.geojson'

    node_state.print_current_state(ingress_dir)

    if config.acsc_cluster_config is not None:
        print('******************')
        print('Cluster input site summaries')
        # If specified cluster sites first.
        from geowatch.mlops import smart_pipeline
        site_clustering = smart_pipeline.SiteClustering(root_dpath=ingress_dir)
        acsc_cluster_config = ub.udict(Yaml.coerce(config.acsc_cluster_config))
        cluster_region_dpath = (ingress_dir / 'clustered_regions').ensuredir()
        cluster_region_fpath = cluster_region_dpath / ('clustered_' + input_region_path.name)
        tocrop_region_fpath = cluster_region_fpath
        acsc_cluster_config['src'] = input_region_path
        acsc_cluster_config['dst_dpath'] = cluster_region_dpath
        acsc_cluster_config['dst_region_fpath'] = cluster_region_fpath
        site_clustering.configure(acsc_cluster_config)

        print('Show stats about cluster inputs:')
        ub.cmd('geowatch site_stats ' + input_region_path, verbose=3, system=True)

        print('Execute clustering')
        cluster_command = site_clustering._raw_command()
        ub.cmd(cluster_command, check=True, verbose=3, system=True)

        print('Show stats about cluster outputs:')
        ub.cmd('geowatch site_stats ' + cluster_region_fpath, verbose=3, system=True)
    else:
        cluster_region_dpath = None
        tocrop_region_fpath = input_region_path

    node_state.print_current_state(ingress_dir)

    ta1_sc_kwcoco_path = ingressed_assets['kwcoco_for_sc']

    align_config_default = ub.udict(Yaml.coerce(ub.codeblock(
        '''
        force_nodata: -9999
        include_channels: "red|green|blue|quality"
        site_summary: True
        geo_preprop: auto
        keep: null
        convexify_regions: True
        target_gsd: 2
        context_factor: 1.5
        force_min_gsd: 4
        img_workers: 1
        aux_workers: 2
        rpc_align_method: affine_warp
        image_timeout: 20minutes
        asset_timeout: 10minutes
        verbose: 1
        ''')))
    align_config = align_config_default | Yaml.coerce(config.acsc_align_config)
    if align_config['aux_workers'] == 'auto':
        align_config['aux_workers'] = align_config['include_channels'].count('|') + 1

    # 4. Crop ingress KWCOCO dataset to region for SC
    print('******************')
    print("* Cropping KWCOCO dataset to region for SC*")
    ta1_sc_cropped_kwcoco_prefilter_path = ingress_dir / 'cropped_kwcoco_for_sc_prefilter.json'
    ta1_sc_cropped_kwcoco_path = ingress_dir / 'cropped_kwcoco_for_sc.json'

    align_config['src'] = ta1_sc_kwcoco_path
    align_config['dst'] = ta1_sc_cropped_kwcoco_prefilter_path
    align_config['regions'] = tocrop_region_fpath
    # Validate align config before running anything
    align_config = coco_align.CocoAlignGeotiffConfig(**align_config)
    print('align_config = {}'.format(ub.urepr(align_config, nl=1)))

    # Not sure if one is more stable than the other, but cmd seems fine and
    # gives us nicer logs. Prefer that one when possible.
    EXEC_MODE = 'cmd'
    if EXEC_MODE == 'import':
        coco_align.main(cmdline=False, **align_config)
    elif EXEC_MODE == 'cmd':
        align_node = ProcessNode(
            command='python -m geowatch.cli.coco_align',
            config=align_config,
        )
        command = align_node.final_command()
        ub.cmd(command, check=True, capture=False, verbose=3, shell=True)
    else:
        raise KeyError(EXEC_MODE)

    # Reroot the kwcoco files to be relative and make it easier to work with
    # downloaded results
    ub.cmd(['kwcoco', 'reroot', f'--src={ta1_sc_cropped_kwcoco_prefilter_path}', '--inplace=1', '--absolute=0'])

    node_state.print_current_state(ingress_dir)

    ### Filter / clean geotiffs (probably should be a separate step)
    CLEAN_GEOTIFFS = 0
    if CLEAN_GEOTIFFS:
        # Detect blocky black regions in geotiffs and switch them to NODATA
        # Modifies geotiffs inplace
        remove_bad_images_node = ProcessNode(
            command='geowatch clean_geotiffs',
            in_paths={
                'src': ta1_sc_cropped_kwcoco_prefilter_path,
            },
            config={
                'prefilter_channels': 'red',
                'channels': 'red|green|blue|nir',
                'workers': 'avail',
                'dry': False,
                'probe_scale': None,
                'nodata_value': -9999,
                'min_region_size': 256,
            },
            node_dpath='.'
        )
        command = remove_bad_images_node.final_command()
        ub.cmd(command, shell=True, capture=False, verbose=3, check=True)

    REMOVE_BAD_IMAGES = 1
    if REMOVE_BAD_IMAGES:
        # Remove images that are nearly all nan
        remove_bad_images_node = ProcessNode(
            command='geowatch remove_bad_images',
            in_paths={
                'src': ta1_sc_cropped_kwcoco_prefilter_path,
            },
            out_paths={
                'dst': ta1_sc_cropped_kwcoco_path,
            },
            config={
                'channels': 'red|green|blue|pan',
                'workers': 'avail',
                'interactive': False,
                'delete_assets': True,
                'overview': 0,
            },
            node_dpath='.'
        )
        command = remove_bad_images_node.final_command()
        print('Using system=True')
        ub.cmd(command, system=True, capture=False, verbose=3, check=True)
    else:
        print('Not removing bad images')
        ta1_sc_cropped_kwcoco_prefilter_path.copy(ta1_sc_cropped_kwcoco_path)

    node_state.print_current_state(ingress_dir)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    sc_cropped_assets_dir = ingress_dir / f'{region_id}'
    sc_cropped_assets_dir.ensuredir()
    # Put a dummy file in this directory so we can upload a nearly-empty folder
    # to S3
    (sc_cropped_assets_dir / 'dummy').write_text('dummy')

    print("* Egressing KWCOCO dataset and associated STAC item *")
    ingressed_assets['cropped_kwcoco_for_sc'] = ta1_sc_cropped_kwcoco_path

    if cluster_region_dpath is not None:
        ingressed_assets['clustered_region_dpath'] = cluster_region_dpath

    ingressed_assets['cropped_kwcoco_for_sc_assets'] = sc_cropped_assets_dir

    # This is an alias to the ACSC dataset and assets that team feature
    # scripts will update.
    assc_kwcoco_rawband_dpath = sc_cropped_assets_dir
    assc_kwcoco_teamfeat_dpath = (ingress_dir / '_teamfeats').ensuredir()
    (assc_kwcoco_rawband_dpath / 'dummy').write_text('dummy')
    (assc_kwcoco_teamfeat_dpath / 'dummy').write_text('dummy')
    ingressed_assets['enriched_acsc_kwcoco_file'] = ta1_sc_cropped_kwcoco_path
    ingressed_assets['enriched_acsc_kwcoco_teamfeats'] = assc_kwcoco_teamfeat_dpath
    ingressed_assets['enriched_acsc_kwcoco_rawbands'] = assc_kwcoco_rawband_dpath

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=config.aws_profile,
                     dryrun=config.dryrun,
                     newline=False)


if __name__ == "__main__":
    main()



"""Collection of general functions and classes."""

import inspect
import os
import re
import stat
import sys
import textwrap
import warnings
from argparse import ArgumentError
from collections import Counter, namedtuple
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from functools import wraps
from importlib import import_module
from io import StringIO
from types import BuiltinFunctionType, FunctionType, ModuleType
from typing import (
    IO,
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    get_type_hints,
)

from ._common import (
    ClassType,
    get_generic_origin,
    is_subclass,
    parser_capture,
    parser_context,
)
from ._deprecated import PathDeprecations
from ._loaders_dumpers import json_dump, load_value
from ._optionals import (
    fsspec_support,
    get_config_read_mode,
    import_fsspec,
    import_requests,
    url_support,
)
from ._type_checking import ArgumentParser

__all__ = [
    "capture_parser",
    "class_from_function",
    "Path",
    "register_unresolvable_import_paths",
]


NoneType = type(None)


default_config_option_help = "Path to a configuration file."


@dataclass
class UrlData:
    scheme: str
    url_path: str


def argument_error(message: str) -> ArgumentError:
    return ArgumentError(None, message)


class JsonargparseWarning(UserWarning):
    pass


def warning(message, category=JsonargparseWarning, stacklevel=1):
    message = textwrap.fill(textwrap.dedent(message), 110).strip()
    warnings.warn(
        "\n" + textwrap.indent(message, "    ") + "\n",
        category=category,
        stacklevel=stacklevel + 1,
    )


class CaptureParserException(Exception):
    def __init__(self, parser: Optional[ArgumentParser]):
        self.parser = parser
        super().__init__("" if parser else "No parse_args call to capture the parser.")


def capture_parser(function: Callable, *args, **kwargs) -> ArgumentParser:
    """Returns the parser object used within the execution of a function.

    The function execution is stopped on the start of the call to parse_args. No
    parsing is done or execution of instructions after the parse_args.

    Args:
        function: A callable that internally creates a parser and calls parse_args.
        *args: Positional arguments used to run the function.
        **kwargs: Keyword arguments used to run the function.

    Raises:
        CaptureParserException: If the function does not call parse_args.
    """
    try:
        with parser_context(parser_capture=True):
            function(*args, **kwargs)
    except CaptureParserException as ex:
        return ex.parser  # type: ignore[return-value]
    raise CaptureParserException(None)


def return_parser_if_captured(parser: ArgumentParser):
    if parser_capture.get():
        raise CaptureParserException(parser)


def identity(value):
    return value


NestedArg = namedtuple("NestedArg", "key val")


def parse_value_or_config(
    value: Any, enable_path: bool = True, simple_types: bool = False
) -> Tuple[Any, Optional["Path"]]:
    """Parses yaml/json config in a string or a path"""
    nested_arg: Union[bool, NestedArg] = False
    if isinstance(value, NestedArg):
        nested_arg = value
        value = nested_arg.val
    cfg_path = None
    if enable_path and type(value) is str and value != "-":
        try:
            cfg_path = Path(value, mode=get_config_read_mode())
        except TypeError:
            pass
        else:
            with cfg_path.relative_path_context():
                value = load_value(cfg_path.get_content(), simple_types=simple_types)
    if type(value) is str and value.strip() != "":
        parsed_val = load_value(value, simple_types=simple_types)
        if type(parsed_val) is not str:
            value = parsed_val
    if isinstance(value, dict) and cfg_path is not None:
        value["__path__"] = cfg_path
    if nested_arg:
        value = NestedArg(key=nested_arg.key, val=value)  # type: ignore[union-attr]
    return value, cfg_path


class CachedStdin(StringIO):
    """Used to allow reading sys.stdin multiple times."""


def get_cached_stdin() -> CachedStdin:
    if not isinstance(sys.stdin, CachedStdin):
        sys.stdin = CachedStdin(sys.stdin.read())
    return sys.stdin


def read_cached_stdin() -> str:
    stdin = get_cached_stdin()
    value = stdin.read()
    stdin.seek(0)
    return value


def import_object(name: str):
    """Returns an object in a module given its dot import path."""
    if not isinstance(name, str) or "." not in name:
        raise ValueError(f"Expected a dot import path string: {name}")
    if not all(x.isidentifier() for x in name.split(".")):
        raise ValueError(f"Unexpected import path format: {name}")
    name_module, name_object = name.rsplit(".", 1)
    try:
        parent = __import__(name_module, fromlist=[name_object])
    except ModuleNotFoundError as ex:
        if "." not in name_module:
            raise ex
        name_module, name_object1 = name_module.rsplit(".", 1)
        parent = getattr(__import__(name_module, fromlist=[name_object1]), name_object1)
    return getattr(parent, name_object)


unresolvable_import_paths = {}


def register_unresolvable_import_paths(*modules: ModuleType):
    """Saves import paths of module objects for which its import path is unresolvable from the object alone.

    Objects with unresolvable import paths have the __module__ attribute set to None.
    """
    for module in modules:
        for val in vars(module).values():
            if (
                getattr(val, "__module__", None) is None
                and getattr(val, "__name__", None)
                and type(val) in {BuiltinFunctionType, FunctionType, Type}
            ):
                unresolvable_import_paths[val] = f"{module.__name__}.{val.__name__}"


def get_module_var_path(module_path: str, value: Any) -> Optional[str]:
    module = import_module(module_path)
    for name, var in vars(module).items():
        if var is value:
            return module_path + "." + name
    return None


def get_import_path(value: Any) -> Optional[str]:
    """Returns the shortest dot import path for the given object."""
    path = None
    value = get_generic_origin(value)
    module_path = getattr(value, "__module__", None)
    qualname = getattr(value, "__qualname__", "")

    if module_path is None:
        path = unresolvable_import_paths.get(value)
        if path:
            module_path, _ = path.rsplit(".", 1)
    elif (not qualname and not inspect.isclass(value)) or (
        inspect.ismethod(value) and not inspect.isclass(value.__self__)
    ):
        path = get_module_var_path(module_path, value)
    elif qualname:
        path = module_path + "." + qualname

    if not path:
        raise ValueError(f"Not possible to determine the import path for object {value}.")

    if qualname and module_path and "." in module_path:
        module_parts = module_path.split(".")
        for num in range(len(module_parts)):
            module_path = ".".join(module_parts[: num + 1])
            module = import_module(module_path)
            if "." in qualname:
                obj_name, attr = qualname.rsplit(".", 1)
                obj = getattr(module, obj_name, None)
                if getattr(obj, attr, None) is value:
                    path = module_path + "." + qualname
                    break
            elif getattr(module, qualname, None) is value:
                path = module_path + "." + qualname
                break
    return path


def object_path_serializer(value):
    try:
        path = get_import_path(value)
        reimported = import_object(path)
        if value is not reimported:
            raise ValueError
        return path
    except Exception as ex:
        raise ValueError(f"Only possible to serialize an importable object, given {value}: {ex}") from ex


def get_typehint_origin(typehint):
    if not hasattr(typehint, "__origin__"):
        typehint_class = get_import_path(typehint.__class__)
        if typehint_class == "types.UnionType":
            return Union
        if typehint_class == "typing._TypedDictMeta":
            return dict
    return getattr(typehint, "__origin__", None)


current_path_dir: ContextVar[Optional[str]] = ContextVar("current_path_dir", default=None)


@contextmanager
def change_to_path_dir(path: Optional["Path"]) -> Iterator[Optional[str]]:
    """A context manager for running code in the directory of a path."""
    path_dir = current_path_dir.get()
    chdir: Union[bool, str] = False
    if path is not None:
        if path._url_data and (path.is_url or path.is_fsspec):
            scheme = path._url_data.scheme
            path_dir = path._url_data.url_path
        else:
            scheme = ""
            path_dir = path.absolute
            chdir = True
        if "d" not in path.mode:
            path_dir = os.path.dirname(path_dir)
        path_dir = scheme + path_dir

    token = current_path_dir.set(path_dir)
    if chdir and path_dir:
        chdir = os.getcwd()
        path_dir = os.path.abspath(path_dir)
        os.chdir(path_dir)

    try:
        yield path_dir
    finally:
        current_path_dir.reset(token)
        if chdir:
            os.chdir(chdir)


def hash_item(item):
    try:
        if isinstance(item, (dict, list)):
            item_hash = hash(json_dump(item))
        else:
            item_hash = hash(item)
    except Exception:
        item_hash = hash(repr(item))
    return item_hash


def unique(iterable):
    unique_items = []
    seen = set()
    for item in iterable:
        key = hash_item(item)
        if key not in seen:
            unique_items.append(item)
            seen.add(key)
    return unique_items


def iter_to_set_str(val, sep=","):
    val = unique(val)
    if len(val) == 1:
        return str(val[0])
    return "{" + sep.join(str(x) for x in val) + "}"


def indent_text(text: str, first_line: bool = True) -> str:
    if first_line:
        return textwrap.indent(text, "  ")
    lines = text.splitlines()
    if len(lines) == 1:
        return text
    return lines[0] + os.linesep + textwrap.indent(os.linesep.join(lines[1:]), "  ")


def get_private_kwargs(data, **kwargs):
    extracted = [data.pop(name, default) for name, default in kwargs.items()]
    if data:
        raise ValueError(f"Unexpected keyword parameters: {set(data.keys())}")
    return extracted[0] if len(extracted) == 1 else extracted


def known_to_fsspec(path: str) -> bool:
    import_fsspec("known_to_fsspec")
    from fsspec.registry import known_implementations

    for protocol in known_implementations:
        if path.startswith(protocol + "://") or path.startswith(protocol + "::"):
            return True
    return False


class ClassFromFunctionBase:
    wrapped_function: Callable


def class_from_function(
    func: Callable[..., ClassType],
    func_return: Optional[Type[ClassType]] = None,
    name: Optional[str] = None,
) -> Type[ClassType]:
    """Creates a dynamic class which if instantiated is equivalent to calling func.

    Args:
        func: A function that returns an instance of a class.
        func_return: The return type of the function. Required if func does not have a return type annotation.
        name: The name of the class. Defaults to function name suffixed with "_class".
    """
    if func_return is None:
        func_return = inspect.signature(func).return_annotation
    if func_return is inspect.Signature.empty:
        raise ValueError(f"{func} does not have a return type annotation")
    if isinstance(func_return, str):
        try:
            func_return = get_type_hints(func)["return"]
            if isinstance(func_return, __import__("typing").ForwardRef):
                func_return = func_return._evaluate(func.__globals__, {})
        except Exception as ex:
            func_return = inspect.signature(func).return_annotation
            raise ValueError(f"Unable to dereference {func_return}, the return type of {func}: {ex}") from ex

    if not name:
        name = func.__qualname__.replace(".", "__") + "_class"

    caller_module = inspect.getmodule(inspect.stack()[1][0]) or inspect.getmodule(class_from_function)
    assert caller_module
    if hasattr(caller_module, name):
        cls = getattr(caller_module, name)
        mro = inspect.getmro(cls) if inspect.isclass(cls) else ()
        if (
            len(mro) > 1
            and mro[1] is func_return
            and is_subclass(cls, ClassFromFunctionBase)
            and cls.wrapped_function is func
            and cls.__name__ == name
        ):
            return cls
        raise ValueError(f"{caller_module.__name__} already defines {name!r}, please use a different name")

    @wraps(func)
    def __new__(cls, *args, **kwargs):
        return func(*args, **kwargs)

    class ClassFromFunction(func_return, ClassFromFunctionBase):  # type: ignore[valid-type,misc]
        pass

    setattr(caller_module, name, ClassFromFunction)
    ClassFromFunction.wrapped_function = func
    ClassFromFunction.__new__ = __new__  # type: ignore[method-assign]
    ClassFromFunction.__doc__ = func.__doc__
    ClassFromFunction.__module__ = caller_module.__name__
    ClassFromFunction.__name__ = name
    ClassFromFunction.__qualname__ = name
    return ClassFromFunction


def parse_url(url: str) -> Optional[UrlData]:
    index = url.rfind("://")
    if index <= 0:
        return None
    return UrlData(
        scheme=url[: index + 3],
        url_path=url[index + 3 :],
    )


def is_absolute_path(path: str) -> bool:
    if path.find("://") > 0:
        return True
    return os.path.isabs(path)


def resolve_relative_path(path: str) -> str:
    parts = path.split("/")
    resolved: List[str] = []
    for part in parts:
        if part == "..":
            resolved.pop()
        elif part != ".":
            resolved.append(part)
    return "/".join(resolved)


class PathError(TypeError):
    """Exception raised for errors in the Path class."""


class Path(PathDeprecations):
    """Stores a (possibly relative) path and the corresponding absolute path.

    The absolute path can be obtained without having to remember the working
    directory (or parent remote path) from when the object was created.

    When a Path instance is created, it is checked that: the path exists,
    whether it is a file or directory and whether it has the required access
    permissions (f=file, d=directory, r=readable, w=writeable, x=executable,
    c=creatable, u=url, s=fsspec or in uppercase meaning not, i.e., F=not-file,
    D=not-directory, R=not-readable, W=not-writeable and X=not-executable).

    The creatable flag "c" can be given one or two times. If give once, the
    parent directory must exist and be writeable. If given twice, the parent
    directory does not have to exist, but should be allowed to create.

    An instance of Path class can also refer to the standard input or output.
    To do that, path must be set with the value "-"; it is a common practice.
    Then, getting the content or opening it will automatically be done on
    standard input or output.
    """

    _url_data: Optional[UrlData]
    _file_scheme = re.compile("^file:///?")

    def __init__(
        self,
        path: Union[str, os.PathLike, "Path"],
        mode: str = "fr",
        cwd: Optional[Union[str, os.PathLike]] = None,
        **kwargs,
    ):
        """Initializer for Path instance.

        Args:
            path: The path to check and store.
            mode: The required type and access permissions among [fdrwxcuFDRWX].
            cwd: Working directory for relative paths. If None, os.getcwd() is used.

        Raises:
            ValueError: If the provided mode is invalid.
            PathError: If the path does not exist or does not agree with the mode.
        """
        self._deprecated_kwargs(kwargs)
        self._check_mode(mode)
        self._std_io = False

        is_url = False
        is_fsspec = False
        if isinstance(path, Path):
            self.std_io = path._std_io
            is_url = path.is_url
            is_fsspec = path.is_fsspec
            url_data = path._url_data
            cwd = path.cwd
            abs_path = path._absolute
            path = path._relative
        elif isinstance(path, (str, os.PathLike)):
            if path == "-":
                self._std_io = True
            path = os.fspath(path)
            cwd = os.fspath(cwd) if cwd else None
            abs_path = os.path.expanduser(path)
            if self._file_scheme.match(abs_path):
                abs_path = self._file_scheme.sub("" if os.name == "nt" else "/", abs_path)
            is_absolute = is_absolute_path(abs_path)
            url_data = parse_url(abs_path)
            cwd_url_data = parse_url(cwd or current_path_dir.get() or os.getcwd())
            if ("u" in mode or "s" in mode) and (url_data or (cwd_url_data and not is_absolute)):
                if cwd_url_data and not is_absolute:
                    abs_path = resolve_relative_path(cwd_url_data.url_path + "/" + path)
                    abs_path = cwd_url_data.scheme + abs_path
                    url_data = parse_url(abs_path)
                if cwd is None:
                    cwd = current_path_dir.get() or os.getcwd()
                if "u" in mode and url_support:
                    is_url = True
                elif "s" in mode and fsspec_support and known_to_fsspec(abs_path):
                    is_fsspec = True
            else:
                if cwd is None:
                    cwd = os.getcwd()
                abs_path = abs_path if is_absolute else os.path.join(cwd, abs_path)
                url_data = None
        else:
            raise PathError("Expected path to be a string, os.PathLike or a Path object.")

        if not self._skip_check and is_url:
            if "r" in mode:
                requests = import_requests("Path with URL support")
                try:
                    requests.head(abs_path).raise_for_status()
                except requests.HTTPError as ex:
                    raise PathError(f"{abs_path} HEAD not accessible :: {ex}") from ex
        elif not self._skip_check and is_fsspec:
            fsspec_mode = "".join(c for c in mode if c in {"r", "w"})
            if fsspec_mode:
                fsspec = import_fsspec("Path")
                try:
                    handle = fsspec.open(abs_path, fsspec_mode)
                    handle.open()
                    handle.close()
                except (FileNotFoundError, KeyError) as ex:
                    raise PathError(f"Path does not exist: {abs_path!r}") from ex
                except PermissionError as ex:
                    raise PathError(f"Path exists but no permission to access: {abs_path!r}") from ex
        elif not self._skip_check and not self._std_io:
            ptype = "Directory" if "d" in mode else "File"
            if "c" in mode:
                pdir = os.path.realpath(os.path.join(abs_path, ".."))
                if not os.path.isdir(pdir) and mode.count("c") == 2:
                    ppdir = None
                    while not os.path.isdir(pdir) and pdir != ppdir:
                        ppdir = pdir
                        pdir = os.path.realpath(os.path.join(pdir, ".."))
                if not os.path.isdir(pdir):
                    raise PathError(f"{ptype} is not creatable since parent directory does not exist: {abs_path!r}")
                if not os.access(pdir, os.W_OK):
                    raise PathError(f"{ptype} is not creatable since parent directory not writeable: {abs_path!r}")
                if "d" in mode and os.access(abs_path, os.F_OK) and not os.path.isdir(abs_path):
                    raise PathError(f"{ptype} is not creatable since path already exists: {abs_path!r}")
                if "f" in mode and os.access(abs_path, os.F_OK) and not os.path.isfile(abs_path):
                    raise PathError(f"{ptype} is not creatable since path already exists: {abs_path!r}")
            elif "d" in mode or "f" in mode:
                if not os.access(abs_path, os.F_OK):
                    raise PathError(f"{ptype} does not exist: {abs_path!r}")
                if "d" in mode and not os.path.isdir(abs_path):
                    raise PathError(f"Path is not a directory: {abs_path!r}")
                if "f" in mode and not (os.path.isfile(abs_path) or stat.S_ISFIFO(os.stat(abs_path).st_mode)):
                    raise PathError(f"Path is not a file: {abs_path!r}")

            if "r" in mode and not os.access(abs_path, os.R_OK):
                raise PathError(f"{ptype} is not readable: {abs_path!r}")
            if "w" in mode and not os.access(abs_path, os.W_OK):
                raise PathError(f"{ptype} is not writeable: {abs_path!r}")
            if "x" in mode and not os.access(abs_path, os.X_OK):
                raise PathError(f"{ptype} is not executable: {abs_path!r}")
            if "D" in mode and os.path.isdir(abs_path):
                raise PathError(f"Path is a directory: {abs_path!r}")
            if "F" in mode and (os.path.isfile(abs_path) or stat.S_ISFIFO(os.stat(abs_path).st_mode)):
                raise PathError(f"Path is a file: {abs_path!r}")
            if "R" in mode and os.access(abs_path, os.R_OK):
                raise PathError(f"{ptype} is readable: {abs_path!r}")
            if "W" in mode and os.access(abs_path, os.W_OK):
                raise PathError(f"{ptype} is writeable: {abs_path!r}")
            if "X" in mode and os.access(abs_path, os.X_OK):
                raise PathError(f"{ptype} is executable: {abs_path!r}")

        self._relative = path
        self._absolute = abs_path
        self._cwd = cwd
        self._mode = mode
        self._is_url = is_url
        self._is_fsspec = is_fsspec
        self._url_data = url_data

    @property
    def relative(self) -> str:
        """Returns the relative representation of the path (how the path was given on instance creation)."""
        return self._relative

    @property
    def absolute(self) -> str:
        """Returns the absolute representation of the path."""
        return self._absolute

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def is_url(self) -> bool:
        return self._is_url

    @property
    def is_fsspec(self) -> bool:
        return self._is_fsspec

    def __str__(self):
        return self._relative

    def __repr__(self):
        name = "Path_" + self._mode
        name = self._repr_skip_check(name)
        cwd = ""
        if self._relative != self._absolute:
            cwd = ", cwd=" + self._cwd
        return f"{name}({self._relative}{cwd})"

    def __fspath__(self) -> str:
        return self._absolute

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Path):
            return self._absolute == other._absolute
        elif isinstance(other, str):
            return str(self) == other
        return False

    def __call__(self, absolute: bool = True) -> str:
        """Returns the path as a string.

        Args:
            absolute: If false returns the original path given, otherwise the corresponding absolute path.
        """
        return self._absolute if absolute else self._relative

    def get_content(self, mode: str = "r") -> str:
        """Returns the contents of the file or the remote path."""
        if self._std_io:
            return read_cached_stdin()
        elif self._is_url:
            assert mode == "r"
            requests = import_requests("Path.get_content")
            response = requests.get(self._absolute)
            response.raise_for_status()
            return response.text
        elif self._is_fsspec:
            fsspec = import_fsspec("Path.get_content")
            with fsspec.open(self._absolute, mode) as handle:
                with handle as input_file:
                    return input_file.read()
        else:
            with open(self._absolute, mode) as input_file:
                return input_file.read()

    @contextmanager
    def open(self, mode: str = "r") -> Iterator[IO]:
        """Return an opened file object for the path."""
        if self._std_io:
            if "r" in mode:
                yield get_cached_stdin()
            elif "w" in mode:
                yield sys.stdout
        elif self._is_url:
            yield StringIO(self.get_content())
        elif self._is_fsspec:
            fsspec = import_fsspec("Path.open")
            with fsspec.open(self._absolute, mode) as handle:
                yield handle
        else:
            with open(self._absolute, mode) as handle:
                yield handle

    @contextmanager
    def relative_path_context(self) -> Iterator[str]:
        """Context manager to use this path's parent (directory or URL) for relative paths defined within."""
        with change_to_path_dir(self) as path_dir:
            assert isinstance(path_dir, str)
            yield path_dir

    @staticmethod
    def _check_mode(mode: str):
        if not isinstance(mode, str):
            raise ValueError("Expected mode to be a string.")
        if len(set(mode) - set("fdrwxcusFDRWX")) > 0:
            raise ValueError("Expected mode to only include [fdrwxcusFDRWX] flags.")
        for flag, count in Counter(mode).items():
            if count > (2 if flag == "c" else 1):
                raise ValueError(f'Too many occurrences ({count}) for flag "{flag}".')
        if "f" in mode and "d" in mode:
            raise ValueError('Both modes "f" and "d" not possible.')
        if "u" in mode and "d" in mode:
            raise ValueError('Both modes "d" and "u" not possible.')
        if "s" in mode and "d" in mode:
            raise ValueError('Both modes "d" and "s" not possible.')




"""
Example:
    from geowatch.tasks.fusion.datamodules.network_io import RGBImageBatchItem
    item0 = RGBImageBatchItem.demo(index=0)
    item1 = RGBImageBatchItem.demo(index=1)
    batch_items = [item0, item1]
    self.imdata_chw.shape
    self.channels

    from geowatch.tasks.fusion.methods.torchvision_efficientnet import *  # NOQA
    self = FCNResNet50()
    batch = self.collate(batch_items)
    out = self.forward(batch)

    self = EfficientNetB7()
    out = self.forward(batch)

"""
import torch
import pytorch_lightning as pl
# from torch.nn import functional as F
from geowatch.tasks.fusion.methods import heads as heads_module
from geowatch.tasks.fusion.methods.watch_module_mixins import WatchModuleMixins
from geowatch.utils.util_netharn import InputNorm, Identity
import ubelt as ub  # NOQA
from geowatch.tasks.fusion.datamodules import network_io
from typing import Iterator  # NOQA
# from geowatch.tasks.fusion.methods.loss import coerce_criterion
# from geowatch.tasks.fusion.methods.heads import TaskHeads


try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


class TorchvisionWrapper(pl.LightningModule):

    # def __init__(self, heads):
    #     super().__init__()
    #     self.ots_model = self.define_ots_model()
    #     self.heads = heads_module.TaskHeads(heads)

    def define_ots_model(self):
        """
        This should define the backbone model.
        """
        raise NotImplementedError('Child class must define this')

    def forward(self, batch):
        imdata_bchw = batch['imdata_bchw']
        # self.ots_model.features.forward(imdata_bchw).shape
        out = self.ots_model.forward(imdata_bchw)
        return out

    def collate(self, batch_items):
        imdatas = [batch_item.imdata_chw for batch_item in batch_items]
        imdata_bchw = torch.stack(imdatas)
        nonlocal_class_ohes = [batch_item.nonlocal_class_ohe for batch_item in batch_items]
        nonlocal_class_ohe = torch.stack(nonlocal_class_ohes)
        batch = {
            'imdata_bchw': imdata_bchw,
            'nonlocal_class_ohe': nonlocal_class_ohe,
        }
        return batch


class TorchvisionSegmentationWrapper(TorchvisionWrapper):
    ...


class TorchvisionClassificationWrapper(TorchvisionWrapper):
    ...


class TorchvisionDetectionWrapper(TorchvisionWrapper):
    ...


class EfficientNetB7(TorchvisionClassificationWrapper):
    def define_ots_model(self):
        import torchvision
        ots_model = torchvision.models.efficientnet_b7(weights='EfficientNet_B7_Weights.IMAGENET1K_V1')
        # adaptpool = ots_model.avgpool  # adaptive
        # head = ots_model.classifier  # 0x1000 classifier
        # out_feat = ots_model.features[-1]  # 640x2560 output features.
        # backbone = ots_model.features[0:]  # 640x2560 output features.
        # stem = ots_model.features[0]  # 3x64 input stem
        return ots_model


class FCNResNet50(TorchvisionSegmentationWrapper, WatchModuleMixins):
    """
    Ignore:
        from geowatch.tasks.fusion.datamodules.network_io import RGBImageBatchItem
        item1 = RGBImageBatchItem.demo()
        item2 = RGBImageBatchItem.demo()
        batch_items = [item1, item2]
        from geowatch.tasks.fusion.methods.torchvision_nets import *  # NOQA
        heads = ub.codeblock(
            '''
            feat_dim: 2048
            tasks:
                # Mirrors the simple FCNHead in torchvision
                - name: saliency
                  type: mlp
                  hidden_channels: [256]
                  out_channels: 2
                  dropout: 0.1
                  norm: batch
                  loss:
                      type: focal
                      gamma: 2.0
                  global_weight: 1.0
            ''')
        self = FCNResNet50(
            heads=head_text,
        )
        batch = self.collate(batch_items)

    Ignore:
        >>> # Test with datamodule
        >>> import ubelt as ub
        >>> from geowatch.tasks.fusion import datamodules
        >>> datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        >>>     train_dataset='special:vidshapes8', chip_size=32,
        >>>     batch_size=1, time_steps=1, num_workers=2, normalize_inputs=10)
        >>> datamodule.setup('fit')
        >>> dataset_stats = datamodule.torch_datasets['train'].cached_dataset_stats(num=3)
        >>> classes = datamodule.torch_datasets['train'].classes
        >>> # Use one of our fusion.architectures in a test
        >>> self = FCNResNet50(
        >>>     classes=classes,
        >>>     dataset_stats=dataset_stats, input_sensorchan=datamodule.input_sensorchan)

    """
    def __init__(self, heads, classes=10, dataset_stats=None):
        super().__init__()
        self.ots_model = self.define_ots_model()
        feat_dim = self.ots_model.backbone.layer4[2].conv3.out_channels
        self.automatic_optimization = True
        assert feat_dim == 2048, 'hard coded sanity check'

        # import kwcoco
        # self.classes = kwcoco.CategoryTree.coerce(classes)
        self.num_classes = len(self.classes)
        self.heads = heads_module.TaskHeads(heads)

    def define_ots_model(self):
        import torchvision
        ots_model = torchvision.models.segmentation.fcn_resnet50(weights='FCN_ResNet50_Weights.COCO_WITH_VOC_LABELS_V1')
        # num_classes = len(self.classes)
        # torchvision.models.segmentation.FCNHead(2048, num_classes)
        return ots_model

    def forward(self, batch):
        # imdata_bchw = batch['imdata_bchw']
        # input_hw = imdata_bchw.shape[-2:]
        # self.ots_model.features.forward(imdata_bchw).shape
        # downscaled_feats = self.ots_model.backbone.forward(imdata_bchw)['out']
        # downscaled_task_outs = self.heads(downscaled_feats)
        raise NotImplementedError

        # x = F.interpolate(downscaled_feats, size=input_hw, mode="bilinear", align_corners=False)
        # out = self.ots_model.forward(imdata_bchw)['out']
        # return out

    def save_package(self, package_path, verbose=1):
        self._save_package(package_path, verbose=verbose)

    def forward_step(self, batch, batch_idx=None, with_loss=True):
        raise NotImplementedError
        # outputs = {
        #     "change_probs": [
        #         [
        #             0.5 * torch.ones(*frame["output_dims"])
        #             for frame in example["frames"]
        #             if frame["change"] is not None
        #         ]
        #         for example in batch
        #     ],
        #     "saliency_probs": [
        #         [
        #             torch.ones(*frame["output_dims"], 2).sigmoid()
        #             for frame in example["frames"]
        #         ]
        #         for example in batch
        #     ],
        #     "class_probs": [
        #         [
        #             torch.ones(*frame["output_dims"], self.num_classes).softmax(dim=-1)
        #             for frame in example["frames"]
        #         ]
        #         for example in batch
        #     ],
        # }

        # if with_loss:
        #     outputs["loss"] = self.dummy_param
        # return outputs


class Resnet50(TorchvisionClassificationWrapper, WatchModuleMixins):
    """
    Ignore:
        >>> import ubelt as ub
        >>> from geowatch.tasks.fusion import datamodules
        >>> datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        >>>     train_dataset='special:vidshapes8', chip_size=32,
        >>>     batch_size=1, time_steps=1, num_workers=2, normalize_inputs=10)
        >>> datamodule.setup('fit')
        >>> dataset_stats = datamodule.torch_datasets['train'].cached_dataset_stats(num=3)
        >>> dataset = datamodule.torch_datasets['train']
        >>> classes = dataset.predictable_classes
        >>> dataset.requested_tasks['nonlocal_class'] = True
        >>> item1 = dataset[0]
        >>> item2 = dataset[1]
        >>> batch_items = [item1, item2]
        >>> from geowatch.tasks.fusion.methods.torchvision_nets import *  # NOQA
        >>> # Use one of our fusion.architectures in a test
        >>> heads = ub.codeblock(
        >>>     '''
        >>>         # Mirrors the simple FCNHead in torchvision
        >>>         - name: nonlocal_class
        >>>           type: mlp
        >>>           hidden_channels: [256]
        >>>           out_channels: 4
        >>>           dropout: 0.1
        >>>           norm: batch
        >>>           loss:
        >>>               type: focal
        >>>               gamma: 2.0
        >>>           head_weight: 1.0
        >>>     ''')
        >>> self = Resnet50(
        >>>     heads=heads,
        >>>     classes=classes,
        >>>     dataset_stats=dataset_stats)
        >>> outputs = self.forward_step(batch_items, with_loss=True)
        >>> canvas = datamodule.draw_batch(batch_items, outputs=outputs)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(canvas)
        >>> kwplot.show_if_requested()

    """

    def __init__(self, heads, classes=None, dataset_stats=None):
        super().__init__()

        input_stats = self.set_dataset_specific_attributes(None, dataset_stats)
        assert len(input_stats) == 1
        stats = list(input_stats.values())[0]
        norm_kw = {'mean': stats['mean'][None, ...],
                   'std': stats['std'][None, ...]}
        self.input_norms = InputNorm(**norm_kw)
        self.ots_model = ots_model = self.define_ots_model()
        feat_dim = ots_model.layer4[-1].conv3.out_channels

        import kwcoco
        self.classes = kwcoco.CategoryTree.coerce(classes)
        self.num_classes = len(self.classes)

        assert feat_dim == 2048, 'hard coded sanity check'
        self.heads = heads_module.TaskHeads(heads, feat_dim=feat_dim)
        self.automatic_optimization = True

    def forward(self, batch):
        imdata_bchw = batch['imdata_bchw']

        # Mean/Std Normalize the Input Batch
        imdata_bchw = self.input_norms(imdata_bchw)
        imdata_bchw.nan_to_num_()

        # Compute backbone features
        feats = self.ots_model(imdata_bchw)

        outputs = self.heads(feats)
        return outputs

    def forward_step(self, batch, with_loss=False, stage='unspecified'):

        if stage == 'train':
            if not self.automatic_optimization:
                # Do we have to do this ourselves?
                # https://lightning.ai/docs/pytorch/stable/common/optimization.html
                opt = self.optimizers()
                opt.zero_grad()

        batch_size = len(batch['imdata_bchw'])
        outputs = self.forward(batch)

        if with_loss:
            losses = self.heads.compute_loss(outputs, batch)
            outputs.update(losses)
            total_loss = losses['loss']
            self.log(f'{stage}_loss', total_loss, prog_bar=True, batch_size=batch_size)

        outputs = network_io.CollatedNetworkOutputs(outputs)

        if stage == 'train':
            if not self.automatic_optimization:
                loss = outputs['loss']
                self.manual_backwards(loss)
        return outputs

    def define_ots_model(self):
        import torchvision
        ots_model = torchvision.models.resnet50(weights='ResNet50_Weights.IMAGENET1K_V2')
        # Hack off the head
        ots_model.fc = Identity()
        return ots_model

    def _to_collated(self, batch_items):
        from geowatch.tasks.fusion.datamodules import network_io
        self._cpu_batch_items = batch_items
        batch_items = network_io.UncollatedRGBImageBatch.from_items(batch_items)
        batch = batch_items.collate()
        batch = batch.to(self.device)
        return batch

    def _grab_batch_from_dataloader(self, dataloader_iter):
        raw_item = next(dataloader_iter)
        batch_items = raw_item
        return self._to_collated(batch_items)

    # These train / vali / test specific methods should be moved to a mixin

    # def training_step(self, dataloader_iter: Iterator) -> None:
    #     self._DataLoaderIterDataFetcher_training_step(dataloader_iter)

    # def training_step(self, batch, batch_idx=None):
    #     return self._PrefetchDataFetcher_training_step(batch, batch_idx)

    # def _DataLoaderIterDataFetcher_training_step(self, dataloader_iter) -> None:
    #     # it is the user responsibility to fetch and move the batch to the right device.
    #     # batch, batch_idx, dataloader_idx
    #     self._PrefetchDataFetcher_training_step(batch)

    # def training_step(self, dataloader_iter: Iterator) -> None:
        # self._grab_batch_from_dataloader()
        # batch = self._to_collated(next(dataloader_iter))
    def training_step(self, batch):
        outputs = self.forward_step(batch, with_loss=True, stage='train')
        return outputs

    # # def validation_step(self, batch, batch_idx=None):
    # def validation_step(self, dataloader_iter: Iterator) -> None:
    #     batch = self._to_collated(next(dataloader_iter))
    #     outputs = self.forward_step(batch, with_loss=True, stage='val')
    #     return outputs

    # # def test_step(self, batch, batch_idx=None):
    # def test_step(self, dataloader_iter: Iterator) -> None:
    #     batch = self._to_collated(next(dataloader_iter))
    #     outputs = self.forward_step(batch, with_loss=True, stage='test')
    #     return outputs

    @profile
    def on_before_batch_transfer(self, batch_items, dataloader_idx):
        from geowatch.tasks.fusion.datamodules import network_io
        self._cpu_batch_items = batch_items
        batch_items = network_io.UncollatedRGBImageBatch.from_items(batch_items)
        batch = batch_items.collate()
        return batch

    # def on_after_batch_transfer(batch, dataloader_idx):
    #     ...

    # def transfer_batch_to_device(batch, device, dataloader_idx):
    #     ...



import kornia
import torch
import torch_lydorn.torchvision
from frame_field_learning import measures
import cv2 as cv
import numpy as np


def compute_distance_transform(tensor: torch.Tensor) -> torch.Tensor:
    device = tensor.device
    array = tensor.cpu().numpy()
    shape = array.shape
    array = array.reshape(-1, *shape[-2:]).astype(np.uint8)
    dist_trans = np.empty(array.shape, dtype=np.float32)
    for i in range(array.shape[0]):
        dist_trans[i] = cv.distanceTransform(array[i], distanceType=cv.DIST_L2, maskSize=cv.DIST_MASK_5, dstType=cv.CV_64F)
    dist_trans = dist_trans.reshape(shape)
    dist_trans = torch.tensor(dist_trans, device=device)
    return dist_trans


def select_crossfield(all_outputs, final_seg):
    # Choose frame field from the replicate that best matches the final seg interior
    dice_loss = measures.dice_loss(all_outputs["seg"][:, :, 0, :, :], final_seg[None, :, 0, :, :])
    # Get index of the replicate that achieves the min dice_loss (as it's a loss, lower is better)
    indices_best = torch.argmin(dice_loss, dim=0)
    batch_range = torch.arange(all_outputs["seg"].shape[1])  # batch size
    # For each batch select frame field from the replicate in indices_best
    final_crossfield = all_outputs["crossfield"][indices_best, batch_range]
    return final_crossfield


def aggr_mean(all_outputs):
    final_outputs = {}
    if "seg" in all_outputs:
        final_seg = torch.mean(all_outputs["seg"], dim=0)
        final_outputs["seg"] = final_seg  # Final seg is between min and max: positive pixels are closer to min
        if "crossfield" in all_outputs:
            final_outputs["crossfield"] = select_crossfield(all_outputs, final_seg)
        '''
        if "depth" in all_outputs:
            # final_depth = torch.mean(all_outputs["depth"], dim=0)
            # final_outputs["depth"] = final_depth
            final_outputs["depth"] = all_outputs["depth"]
        if "shadow" in all_outputs:
            final_shadow = torch.mean(all_outputs["shadow"], dim=0)
            final_outputs["shadow"] = final_shadow
        if "facade" in all_outputs:
            final_facade = torch.mean(all_outputs["facade"], dim=0)
            final_outputs["facade"] = final_facade
        '''
    else:
        raise NotImplementedError("Test Time Augmentation requires segmentation to be computed.")

    if "depth" in all_outputs:
        final_depth = torch.mean(all_outputs["depth"], dim=0)
        final_outputs["depth"] = final_depth

    if "height" in all_outputs:
        final_height = torch.mean(all_outputs["height"], dim=0)
        final_outputs["height"] = final_height

    if "mag" in all_outputs:
        final_mag = torch.mean(all_outputs["mag"], dim=0)
        final_outputs["mag"] = final_mag

    return final_outputs


def aggr_median(all_outputs):
    final_outputs = {}
    if "seg" in all_outputs:
        final_seg, _ = torch.median(all_outputs["seg"], dim=0)
        final_outputs["seg"] = final_seg  # Final seg is between min and max: positive pixels are closer to min
        if "crossfield" in all_outputs:
            final_outputs["crossfield"] = select_crossfield(all_outputs, final_seg)
    else:
        raise NotImplementedError("Test Time Augmentation requires segmentation to be computed.")
    return final_outputs


def aggr_dist_trans(all_outputs, seg_threshold):
    final_outputs = {}
    if "seg" in all_outputs:
        min_seg, _ = torch.min(all_outputs["seg"], dim=0)
        max_seg, _ = torch.max(all_outputs["seg"], dim=0)
        # Final seg will be between min and max seg. The idea is that we don't loose the sharp corners (which taking the mean does)
        dist_ext_to_min_seg = compute_distance_transform(min_seg < seg_threshold)
        dist_int_to_max_seg = compute_distance_transform(seg_threshold < max_seg)
        final_seg = dist_ext_to_min_seg < dist_int_to_max_seg
        final_outputs["seg"] = final_seg  # Final seg is between min and max: positive pixels are closer to min
        if "crossfield" in all_outputs:
            final_outputs["crossfield"] = select_crossfield(all_outputs, final_seg)
    else:
        raise NotImplementedError("Test Time Augmentation requires segmentation to be computed.")
    return final_outputs


def aggr_translated(all_outputs, seg_threshold, image_display=None):
    final_outputs = {}
    if "seg" in all_outputs:
        # Cleanup all_seg by multiplying with the mean seg
        all_seg = all_outputs["seg"]
        all_seg_mask: torch.Tensor = seg_threshold < all_seg
        mean_seg = torch.mean(all_seg_mask.float(), dim=0)
        mean_seg_mask = seg_threshold < mean_seg  # NOQA
        all_cleaned_seg = all_seg_mask * mean_seg[None, ...]
        # all_cleaned_seg_mask = seg_threshold < all_cleaned_seg
        # all_cleaned_seg[~all_cleaned_seg_mask] = 0  # Put 0 where seg is below threshold

        # # --- DEBUG SAVE
        # image_seg_display = plot_utils.get_tensorboard_image_seg_display(image_display, mean_seg)
        # image_seg_display = image_seg_display[0].cpu().detach().numpy().transpose(1, 2, 0)
        # skimage.io.imsave(f"image_seg_display_mean_seg.png", image_seg_display)
        # for i, cleaned_seg in enumerate(all_cleaned_seg):
        #     image_seg_display = plot_utils.get_tensorboard_image_seg_display(image_display, cleaned_seg)
        #     image_seg_display = image_seg_display[0].cpu().detach().numpy().transpose(1, 2, 0)
        #     skimage.io.imsave(f"image_seg_display_replicate_cleaned_{i}.png", image_seg_display)
        # # ---

        # Compute barycenter of all cleaned segs
        range_x = torch.arange(all_cleaned_seg.shape[4], device=all_cleaned_seg.device)
        range_y = torch.arange(all_cleaned_seg.shape[3], device=all_cleaned_seg.device)
        grid_y, grid_x = torch.meshgrid([range_x, range_y])
        grid_xy = torch.stack([grid_x, grid_y], dim=-1)

        # Average of coordinates, weighted by segmentation confidence
        spatial_mean_xy = torch.sum(grid_xy[None, None, None, :, :, :] * all_cleaned_seg[:, :, :, :, :, None], dim=(3, 4)) / torch.sum(all_cleaned_seg[:, :, :, :, :, None], dim=(3, 4))
        # Median of all replicate's means
        median_spatial_mean_xy, _ = torch.median(spatial_mean_xy, dim=0)
        # Compute shift between each replicates and the average
        shift_xy = median_spatial_mean_xy[None, :, :, :] - spatial_mean_xy
        shift_xy *= 2  # The shift for the original segs is twice the shift between cleaned segs (assuming homogenous shifts and enough segs)
        shift_xy = shift_xy.view(-1, shift_xy.shape[-1])
        shape = all_outputs["seg"].shape
        shifted_seg = kornia.geometry.translate(all_outputs["seg"].view(-1, *shape[-3:]), shift_xy).view(shape)

        # # --- DEBUG SAVE
        # for i, replicate_shifted_seg in enumerate(shifted_seg):
        #     image_seg_display = plot_utils.get_tensorboard_image_seg_display(image_display, replicate_shifted_seg)
        #     image_seg_display = image_seg_display[0].cpu().detach().numpy().transpose(1, 2, 0)
        #     skimage.io.imsave(f"image_seg_display_replicate_shifted_{i}.png", image_seg_display)
        # # ---

        # Compute mean shifted seg
        mean_shifted_seg = torch.mean(shifted_seg, dim=0)
        # Select replicate seg (and crossfield) that best matches mean_shifted_seg
        dice_loss = measures.dice_loss(shifted_seg[:, :, 0, :, :], mean_shifted_seg[None, :, 0, :, :])
        # Get index of the replicate that achieves the min dice_loss (as it's a loss, lower is better)
        indices_best = torch.argmin(dice_loss, dim=0)
        batch_range = torch.arange(all_outputs["seg"].shape[1])  # batch size
        # For each batch select seg and frame field from the replicate in indices_best
        final_outputs["seg"] = shifted_seg[indices_best, batch_range]
        if "crossfield" in all_outputs:
            final_outputs["crossfield"] = all_outputs["crossfield"][indices_best, batch_range]

        # if "crossfield" in all_outputs:
        #     final_outputs["crossfield"] = select_crossfield(all_outputs, final_seg)
    else:
        raise NotImplementedError("Test Time Augmentation requires segmentation to be computed.")
    return final_outputs


def tta_inference(model, xb, seg_threshold):

    # Perform inference several times with transformed input image and aggregate results
    replicates = 4 * 2  # 4 rotations, each with vflip/no vflip

    # Init results tensors
    notrans_outputs = model.inference(xb["image"])

    output_keys = notrans_outputs.keys()
    all_outputs = {}
    for key in output_keys:

        if key == 'xydir' or key == 'scale':
            continue

        all_outputs[key] = torch.empty((replicates, *notrans_outputs[key].shape),
                                       dtype=notrans_outputs[key].dtype,
                                       device=notrans_outputs[key].device)
        all_outputs[key][0] = notrans_outputs[key]

    # Flip image
    flipped_image = kornia.geometry.transform.vflip(xb["image"])
    flipped_outputs = model.inference(flipped_image)

    for key in output_keys:

        if key == 'xydir' or key == 'scale':
            continue

        reversed_output = kornia.geometry.transform.vflip(flipped_outputs[key])
        all_outputs[key][1] = reversed_output

    # --- Apply transforms one by one and add results to all_outputs
    for k in range(1, 4):
        rotated_image = torch.rot90(xb["image"], k=k, dims=(-2, -1))
        rotated_outputs = model.inference(rotated_image)
        for key in output_keys:

            if key == 'xydir' or key == 'scale':
                continue

            reversed_output = torch.rot90(rotated_outputs[key], k=-k, dims=(-2, -1))
            if key == "crossfield":
                angle = -k * 90
                # TODO: use a faster implementation of rotate_framefield that only handles angles [0, 90, 180, 270]
                reversed_output = torch_lydorn.torchvision.transforms.functional.rotate_framefield(reversed_output,
                                                                                                   angle)
            all_outputs[key][2 * k] = reversed_output

        # Flip rotated image
        flipped_rotated_image = kornia.geometry.transform.vflip(rotated_image)
        flipped_rotated_outputs = model.inference(flipped_rotated_image)
        for key in output_keys:

            if key == 'xydir' or key == 'scale':
                continue

            reversed_output = torch.rot90(kornia.geometry.transform.vflip(flipped_rotated_outputs[key]), k=-k,
                                          dims=(-2, -1))
            if key == "crossfield":
                angle = -k * 90
                reversed_output = torch_lydorn.torchvision.transforms.functional.vflip_framefield(reversed_output)
                reversed_output = torch_lydorn.torchvision.transforms.functional.rotate_framefield(reversed_output,
                                                                                                   angle)
            all_outputs[key][2 * k + 1] = reversed_output

    # --- DEBUG
    # all_outputs["seg"] *= 0
    # for i in range(all_outputs["seg"].shape[0]):
    #     center = 512
    #     size = 100
    #     shift_x = random.randint(-20, 20)
    #     shift_y = random.randint(-20, 20)
    #     all_outputs["seg"][i][..., center + shift_y - size:center + shift_y + size, center + shift_x - size:center + shift_x + size] = 1
    #     # Add noise
    #     noise_center_x = random.randint(100, 1024-100)
    #     noise_center_y = random.randint(100, 1024-100)
    #     noise_size = 10
    #     all_outputs["seg"][i][..., noise_center_y - noise_size:noise_center_y + noise_size, noise_center_x - noise_size:noise_center_x + noise_size] = 1
    #     # Add more noise
    #     all_outputs["seg"][i] += 0.25 * torch.rand(all_outputs["seg"][i].shape, device=all_outputs["seg"][i].device)
    #     all_outputs["seg"][i] = torch.clamp(all_outputs["seg"][i], 0, 1)

    # # --- DEBUG SAVE
    # image_display = torch_lydorn.torchvision.transforms.functional.batch_denormalize(xb["image"],
    #                                                                                  xb["image_mean"],
    #                                                                                  xb["image_std"])
    # for i, replicate_seg in enumerate(all_outputs["seg"]):
    #     image_seg_display = plot_utils.get_tensorboard_image_seg_display(image_display, replicate_seg)
    #     image_seg_display = image_seg_display[0].cpu().detach().numpy().transpose(1, 2, 0)
    #     skimage.io.imsave(f"image_seg_display_replicate_{i}.png", image_seg_display)
    # # ---

    # --- Aggregate results
    # final_outputs = aggr_dist_trans(all_outputs, seg_threshold)
    # final_outputs = aggr_translated(all_outputs, seg_threshold, image_display=image_display)
    # final_outputs = aggr_translated(all_outputs, seg_threshold)

    # pdb.set_trace()

    final_outputs = aggr_mean(all_outputs)

    # final_outputs['depth'] = notrans_outputs['depth']
    final_outputs['shadow'] = notrans_outputs['shadow']
    final_outputs['facade'] = notrans_outputs['facade']

    final_outputs['xydir'] = notrans_outputs['xydir']
    final_outputs['scale'] = notrans_outputs['scale']

    # final_outputs = aggr_median(all_outputs)

    # # --- DEBUG SAVE
    # image_seg_display = plot_utils.get_tensorboard_image_seg_display(image_display, final_outputs["seg"])
    # image_seg_display = image_seg_display[0].cpu().detach().numpy().transpose(1, 2, 0)
    # skimage.io.imsave("image_seg_display_final.png", image_seg_display)
    # # ---

    # input("Press <Enter>...")

    return final_outputs



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements relative [1, 2, 3] and global [3, 4] positional encodings.

Our Axial-Deeplab [1] proposes position-sensitive self-attention which uses
relative positional encodings for query, key, and value.

[1] Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
    ECCV 2020 Spotlight.
      Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
      Liang-Chieh Chen.
[2] Self-Attention with Relative Position Representations, NAACL 2018.
      Peter Shaw, Jakob Uszkoreit, Ashish Vaswani.
[3] Tensor2Tensor for Neural Machine Translation, arXiv 2018,
    http://arxiv.org/abs/1803.07416.
      Ashish Vaswani, Samy Bengio, Eugene Brevdo, Francois Chollet,
      Aidan N. Gomez, Stephan Gouws, Llion Jones, Łukasz Kaiser,
      Nal Kalchbrenner, Niki Parmar, Ryan Sepassi, Noam Shazeer,
      Jakob Uszkoreit.
[4] Attention Is All You Need, NeurIPS 2017.
      Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones,
      Aidan N. Gomez, Lukasz Kaiser, Illia Polosukhin.
[5] An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale,
    ICLR 2021.
      Alexey Dosovitskiy, Lucas Beyer, Alexander Kolesnikov, Dirk Weissenborn,
      Xiaohua Zhai, Thomas Unterthiner, Mostafa Dehghani, Matthias Minderer,
      Georg Heigold, Sylvain Gelly, Jakob Uszkoreit, Neil Houlsby.
"""

import tensorflow as tf
from absl import logging

# MAX_SPAN defines the maximum shape of positional encoding. It is set as a
# large constant so that we can easily load and use models with global or
# different local spans, but it should not be too large so that it takes a
# reasonable amount of memory. The value 255 is larger than almost all span
# choices (e.g. 65 for local attention, 129, 193, etc.) so 255 is large enough.
# 257 will be a good choice for gpu, but 255 is more efficient on TPU which pads
# tensors to 128x.
MAX_SPAN = 255


def _compute_relative_distance_matrix(query_length, key_length):
    """Computes a relative distance matrix between queries and keys.

    We assume that the queries and the keys are centered, i.e.,
    key_length = memory_flange + query_length + memory_flange.

    The function is based on the _generate_relative_positions_matrix function in
    common_attention.py of tensor2tensor codebase:
    https://github.com/tensorflow/tensor2tensor/blob/5623deb79cfcd28f8f8c5463b58b5bd76a81fd0d/tensor2tensor/layers/common_attention.py#L1670

    Args:
      query_length: An integer, the length of queries.
      key_length: An integer, the length of keys.

    Returns:
      distance_matrix: A [query_length, key_length] tensor.

    Raises:
      ValueError: If (key_length - query_length) is odd, i.e., the assumption does
        not hold.
    """
    if (key_length - query_length) % 2:
        raise ValueError('Key_length should be query_length + 2 * memory_flange.')
    key_index = tf.range(key_length)
    query_index = tf.range(query_length) + (key_length - query_length) // 2
    distance_matrix = key_index[None, :] - query_index[:, None]
    # Shift the distance_matrix so that it is >= 0. Each entry of the
    # distance_matrix distance will index a relative positional embedding.
    distance_matrix = distance_matrix + MAX_SPAN - 1
    if query_length + (key_length - query_length) // 2 > MAX_SPAN:
        logging.warn(
            'Axial attention span is larger than MAX_SPAN. In this case, we use a '
            'single shared embedding for all positions beyond this relative '
            'distance. Please make sure, this behavior is intended.')
        distance_matrix = tf.clip_by_value(distance_matrix, 0, MAX_SPAN * 2 - 2)
    return distance_matrix


class RelativePositionalEncoding(tf.keras.layers.Layer):
    """Generates relative positional encoding.

    The function is based on the _generate_relative_positions_embeddings function
    in common_attention.py of tensor2tensor codebase:
    https://github.com/tensorflow/tensor2tensor/blob/5623deb79cfcd28f8f8c5463b58b5bd76a81fd0d/tensor2tensor/layers/common_attention.py#L1691
    """

    def __init__(self, query_length, key_length, depth, name,
                 initialization_std=1.0, conv_kernel_weight_decay=0.0):
        """Initializes a relative position encoding layer.

        Args:
          query_length: An integer, the length of queries.
          key_length: An integer, the length of keys.
          depth: An integer, the number of embedding channels per head.
          name: A string, the name of the embedding.
          initialization_std: A float, the initialization std for the embedding.
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.

        Returns:
          output: A [query, key, depth] tensor, the relative positional
            encodings for each head and each query-key-pair.
        """
        super(RelativePositionalEncoding, self).__init__(name=name)
        self._initializer = tf.keras.initializers.TruncatedNormal(
            stddev=initialization_std)
        self._regularizer = tf.keras.regularizers.l2(conv_kernel_weight_decay)

        self._relative_distance_matrix = _compute_relative_distance_matrix(
            query_length, key_length)
        self._embedding_shape = (MAX_SPAN * 2 - 1, depth)

    def build(self, input_shape):
        """Builds the embedding weight."""
        del input_shape
        self._embeddings = self.add_weight(
            shape=self._embedding_shape,
            initializer=self._initializer, trainable=True,
            name='embeddings',
            regularizer=self._regularizer)

    def call(self, inputs):
        """A forward pass that gathers the relative positional encoding."""
        del inputs
        # Gather the embeddings according to the relative distances.
        return tf.gather(self._embeddings, self._relative_distance_matrix)


class AddAbsolutePositionalEncoding(tf.keras.layers.Layer):
    """Adds a learnable absolute positional encoding to the input feature.

    Supports both 1D and 2D versions of the positional encoding: (1) 1D positional
    encoding represents each row index with an embedding, and represents each
    column index with another embedding. This results in a total of (height +
    width) learnable embedding vectors. (2) 2D positional encoding adds
    independent embeddings to each input grid position. This choice uses a total
    of (height * width) learnable embedding vectors.
    """

    def __init__(self, name, positional_encoding_type=None,
                 bn_layer=tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay=0.0):
        """Initializes an AddAbsolutePositionEmbedding layer.

        Args:
          name: A string specifying the name of the layer.
          positional_encoding_type: A string, type of the positional encoding.
            Support '2D', '1D', 'none', and None. The feature is returned as is if
            positional_encoding_type is 'none' or None.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.

        Raises:
          ValueError: If positional_encoding_type is not one of '1D', '2D', 'none',
            and None.
        """
        super(AddAbsolutePositionalEncoding, self).__init__(name=name)
        if not any([positional_encoding_type is None,
                    positional_encoding_type.lower() == 'none',
                    positional_encoding_type.lower() == '2d',
                    positional_encoding_type.lower() == '1d']):
            raise ValueError(positional_encoding_type + ' is not supported.')
        self._positional_encoding_type = positional_encoding_type
        # This initialization std is tuned for global attention, but it does not
        # seem to be a sensitive hyper-parameter, since we use batch norm on the
        # positional encodings.
        self._initializer = tf.keras.initializers.TruncatedNormal(stddev=0.2)
        self._kernel_regularizer = tf.keras.regularizers.l2(
            conv_kernel_weight_decay)
        self._bn_layer = bn_layer

    def build(self, input_shape):
        """Builds the layer weights whose shape depends on the 4D input shape."""
        _, height, width, channel = input_shape
        if self._positional_encoding_type.lower() == '2d':
            self._embeddings = self.add_weight(
                shape=(1, height, width, channel),
                initializer=self._initializer, trainable=True,
                name='embeddings',
                regularizer=self._kernel_regularizer)
            self._batch_norm = self._bn_layer(axis=-1, name='batch_norm')
        elif self._positional_encoding_type.lower() == '1d':
            # Generate separable positional encodings for the height axis and the
            # width axis.
            self._height_axis_embeddings = self.add_weight(
                shape=(1, height, 1, channel),
                initializer=self._initializer, trainable=True,
                name='height_axis_embeddings',
                regularizer=self._kernel_regularizer)
            self._height_axis_batch_norm = self._bn_layer(
                axis=-1, name='height_axis_batch_norm')
            self._width_axis_embeddings = self.add_weight(
                shape=(1, height, 1, channel),
                initializer=self._initializer, trainable=True,
                name='width_axis_embeddings',
                regularizer=self._kernel_regularizer)
            self._width_axis_batch_norm = self._bn_layer(
                axis=-1, name='width_axis_batch_norm')

    def call(self, features, training=False):
        """Performs a forward pass.

        Args:
          features: An input [batch, height, width, channels] tensor.
          training: A boolean, whether the model is in training mode.

        Returns:
          output: The sum of the input feature and learnable positional encodings.
        """
        if (self._positional_encoding_type is None or
                self._positional_encoding_type.lower() == 'none'):
            return features
        elif self._positional_encoding_type.lower() == '2d':
            positional_encoding = self._batch_norm(self._embeddings,
                                                   training=training)
        elif self._positional_encoding_type.lower() == '1d':
            height_axis_positional_encoding = self._height_axis_batch_norm(
                self._height_axis_embeddings, training=training)
            width_axis_positional_encoding = self._width_axis_batch_norm(
                self._width_axis_embeddings, training=training)
            positional_encoding = (height_axis_positional_encoding +
                                   width_axis_positional_encoding)
        return features + positional_encoding



"""
TODO: remove this file and use the monkey module directly
"""
from geowatch.monkey import monkey_lightning
disable_lightning_hardware_warnings = monkey_lightning.disable_lightning_hardware_warnings
